#
# Provides Python support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import PythonExporter
from ._ewriter import PythonExpressionWriter, NumPyExpressionWriter

# Importers
# Exporters

_exporters = {
    'python': PythonExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'python': PythonExpressionWriter,
    'numpy': NumPyExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Language keywords
keywords = [
    'and',
    'del',
    'from',
    'not',
    'while',
    'as',
    'elif',
    'global',
    'or',
    'with', 'assert', 'else', 'if', 'pass', 'yield', 'break', 'except',
    'import', 'print', 'class', 'exec', 'in', 'raise', 'continue',
    'finally',
    'is',
    'return',
    'def',
    'for',
    'lambda',
    'try',
]



#!/usr/bin/env python3
#
# Tests the exporters from the format module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import sys
import unittest

import myokit

from myokit.tests import TemporaryDirectory, WarningCollector

debug = False


# Simple test
config_basic = """
[myokit]
[time]
date_format = TEST_DATE_FORMAT
time_format = TEST_TIME_FORMAT
[gui]
[sundials]
lib = one;two
inc = three;four
[opencl]
lib = five;six
inc = three;eight
"""

# Config with extra bits
config_extra = """
[extra]
setting = 12
sotting=2
str=String value
"""

# Config with empty paths and spaces
config_paths = """
[sundials]
lib = five
inc = three;four;
[opencl]
lib = one;;   two point five;three;
inc =
"""


class TestConfig(unittest.TestCase):

    @classmethod
    def setUpClass(cls):

        # Import hidden _config module
        path = sys.path
        try:
            sys.path.append(myokit.DIR_MYOKIT)
            import _config as config
        finally:
            sys.path = path

        # Back-up current settings
        cls._compat_no_capture = myokit.COMPAT_NO_CAPTURE
        cls._compat_no_fd_capture = myokit.COMPAT_NO_FD_CAPTURE
        cls._date_format = myokit.DATE_FORMAT
        cls._time_format = myokit.TIME_FORMAT
        cls._force_pyqt6 = myokit.FORCE_PYQT6
        cls._force_pyqt5 = myokit.FORCE_PYQT5
        cls._force_pyside6 = myokit.FORCE_PYSIDE6
        cls._force_pyside2 = myokit.FORCE_PYSIDE2
        cls._sundials_lib = myokit.SUNDIALS_LIB
        cls._sundials_inc = myokit.SUNDIALS_INC
        cls._opencl_lib = myokit.OPENCL_LIB
        cls._opencl_inc = myokit.OPENCL_INC

        # Temporarily change myokit config dir
        cls._dir_user = myokit.DIR_USER
        cls._temp_dir = TemporaryDirectory()
        cls._temp_dir.__enter__()
        myokit.DIR_USER = cls._temp_dir.path()

        # Config module
        cls._config_module = config

        if debug:
            print('setUpClass completed')

    @classmethod
    def tearDownClass(cls):

        # Reset path
        myokit.DIR_USER = cls._dir_user

        # Reset data and time
        myokit.COMPAT_NO_CAPTURE = cls._compat_no_capture
        myokit.COMPAT_NO_FD_CAPTURE = cls._compat_no_fd_capture
        myokit.DATE_FORMAT = cls._date_format
        myokit.TIME_FORMAT = cls._time_format
        myokit.FORCE_PYQT6 = cls._force_pyqt6
        myokit.FORCE_PYQT5 = cls._force_pyqt5
        myokit.FORCE_PYSIDE6 = cls._force_pyside6
        myokit.FORCE_PYSIDE2 = cls._force_pyside2
        myokit.SUNDIALS_LIB = cls._sundials_lib
        myokit.SUNDIALS_INC = cls._sundials_inc
        myokit.OPENCL_LIB = cls._opencl_lib
        myokit.OPENCL_INC = cls._opencl_inc

        # Reload local settings
        cls._config_module._load()

        # Sanity check
        assert myokit.COMPAT_NO_CAPTURE == cls._compat_no_capture
        assert myokit.COMPAT_NO_FD_CAPTURE == cls._compat_no_fd_capture
        assert myokit.DATE_FORMAT != 'TEST_DATE_FORMAT'
        assert myokit.TIME_FORMAT != 'TEST_TIME_FORMAT'
        assert myokit.FORCE_PYQT6 == cls._force_pyqt6
        assert myokit.FORCE_PYQT5 == cls._force_pyqt5
        assert myokit.FORCE_PYSIDE6 == cls._force_pyside6
        assert myokit.FORCE_PYSIDE2 == cls._force_pyside2
        assert myokit.SUNDIALS_LIB != ['one', 'two']
        assert myokit.SUNDIALS_INC != ['three', 'four']
        assert myokit.OPENCL_LIB != ['five', 'six']
        assert myokit.OPENCL_INC != ['three', 'eight']

        if debug:
            print('tearDownClass completed')

    def test_create(self):
        # Test if the `_create` method works.

        # Test _create
        with TemporaryDirectory() as d:
            filename = d.path('test.ini')
            self._config_module._create(filename)
            self.assertTrue(os.path.isfile(filename))
        self.assertFalse(os.path.isfile(filename))

        # No ini file (calls _create())
        path = self._temp_dir.path('myokit.ini')
        if os.path.exists(path):
            os.remove(path)
        self._config_module._load()
        self.assertTrue(os.path.isfile(path))
        os.remove(path)
        self.assertFalse(os.path.isfile(path))

    def test_load_basic(self):
        # Tests basic entries are loaded.
        myokit.SUNDIALS_LIB = []
        myokit.SUNDIALS_INC = []
        myokit.OPENCL_LIB = []
        myokit.OPENCL_INC = []
        myokit.FORCE_PYSIDE6 = myokit.FORCE_PYSIDE2 = False
        myokit.FORCE_PYQT6 = myokit.FORCE_PYQT5 = False
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write(config_basic)
        self._config_module._load()
        self.assertEqual(myokit.DATE_FORMAT, 'TEST_DATE_FORMAT')
        self.assertEqual(myokit.TIME_FORMAT, 'TEST_TIME_FORMAT')
        self.assertFalse(myokit.FORCE_PYQT6)
        self.assertFalse(myokit.FORCE_PYQT5)
        self.assertFalse(myokit.FORCE_PYSIDE6)
        self.assertFalse(myokit.FORCE_PYSIDE2)
        self.assertEqual(myokit.SUNDIALS_LIB, ['one', 'two'])
        self.assertEqual(myokit.SUNDIALS_INC, ['three', 'four'])
        self.assertEqual(myokit.OPENCL_LIB, ['five', 'six'])
        self.assertEqual(myokit.OPENCL_INC, ['three', 'eight'])

        # Unknown settings are ignored
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write(config_extra)
        self._config_module._load()

    def test_load_empty(self):
        # Test loading with an empty file.

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('')
        self._config_module._load()

    def test_load_paths(self):
        # Test loading of sundials and opencl paths

        # Lists of paths should be filtered for empty values and trimmed
        myokit.SUNDIALS_LIB = []
        myokit.SUNDIALS_INC = []
        myokit.OPENCL_LIB = []
        myokit.OPENCL_INC = []
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write(config_paths)
        self._config_module._load()
        self.assertEqual(myokit.SUNDIALS_LIB, ['five'])
        self.assertEqual(myokit.SUNDIALS_INC, ['three', 'four'])
        self.assertEqual(
            myokit.OPENCL_LIB,
            ['one', 'two point five', 'three'])
        self.assertEqual(myokit.OPENCL_INC, [])

    def test_load_gui_backend(self):
        # Tests setting the GUI back end

        # Qt gui options
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=pyqt6\n')
        self._config_module._load()
        self.assertTrue(myokit.FORCE_PYQT6)
        self.assertFalse(myokit.FORCE_PYQT5)
        self.assertFalse(myokit.FORCE_PYSIDE6)
        self.assertFalse(myokit.FORCE_PYSIDE2)

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=pyqt5\n')
        self._config_module._load()
        self.assertFalse(myokit.FORCE_PYQT6)
        self.assertTrue(myokit.FORCE_PYQT5)
        self.assertFalse(myokit.FORCE_PYSIDE6)
        self.assertFalse(myokit.FORCE_PYSIDE2)

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=pyside6\n')
        self._config_module._load()
        self.assertFalse(myokit.FORCE_PYQT6)
        self.assertFalse(myokit.FORCE_PYQT5)
        self.assertTrue(myokit.FORCE_PYSIDE6)
        self.assertFalse(myokit.FORCE_PYSIDE2)

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=pyside2\n')
        self._config_module._load()
        self.assertFalse(myokit.FORCE_PYQT6)
        self.assertFalse(myokit.FORCE_PYQT5)
        self.assertFalse(myokit.FORCE_PYSIDE6)
        self.assertTrue(myokit.FORCE_PYSIDE2)

        # Empty has no effect
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=\n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertEqual(c.text(), '')
        self.assertFalse(myokit.FORCE_PYQT6)
        self.assertFalse(myokit.FORCE_PYQT5)
        self.assertFalse(myokit.FORCE_PYSIDE6)
        self.assertTrue(myokit.FORCE_PYSIDE2)

        # Invalid raises warning
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[gui]\nbackend=lalala\n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertIn('Expected values for backend are', c.text())
        self.assertTrue(myokit.FORCE_PYSIDE2)

    def test_load_compat_no_capture(self):
        # Tests loading compatibility option: no_capture

        myokit.COMPAT_NO_CAPTURE = False
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_capture = True\n')
        self._config_module._load()
        self.assertTrue(myokit.COMPAT_NO_CAPTURE)

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_capture = False\n')
        self._config_module._load()
        self.assertFalse(myokit.COMPAT_NO_CAPTURE)

        # Empty has no effect
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_capture = \n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertEqual(c.text(), '')
        self.assertFalse(myokit.COMPAT_NO_CAPTURE)
        myokit.COMPAT_NO_CAPTURE = True
        with WarningCollector() as c:
            self._config_module._load()
        self.assertEqual(c.text(), '')
        self.assertTrue(myokit.COMPAT_NO_CAPTURE)

        # Invalid raises warning
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_capture = hiya\n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertTrue(myokit.COMPAT_NO_CAPTURE)
        self.assertIn('Expected values for no_capture are', c.text())

    def test_load_compat_no_capture_fd(self):
        # Tests loading compatibility option: no_fd_capture

        myokit.COMPAT_NO_FD_CAPTURE = False
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_fd_capture = True\n')
        self._config_module._load()
        self.assertTrue(myokit.COMPAT_NO_FD_CAPTURE)

        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_fd_capture = False\n')
        self._config_module._load()
        self.assertFalse(myokit.COMPAT_NO_FD_CAPTURE)

        # Empty has no effect
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_fd_capture = \n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertEqual(c.text(), '')
        self.assertFalse(myokit.COMPAT_NO_FD_CAPTURE)
        myokit.COMPAT_NO_FD_CAPTURE = True
        with WarningCollector() as c:
            self._config_module._load()
        self.assertEqual(c.text(), '')
        self.assertTrue(myokit.COMPAT_NO_FD_CAPTURE)

        # Invalid raises warning
        with open(self._temp_dir.path('myokit.ini'), 'w') as f:
            f.write('[compatibility]\nno_fd_capture = hiya\n')
        with WarningCollector() as c:
            self._config_module._load()
        self.assertTrue(myokit.COMPAT_NO_FD_CAPTURE)
        self.assertIn('Expected values for no_fd_capture are', c.text())


if __name__ == '__main__':
    unittest.main(verbosity=2)



<?
# sim.py
# A pype template for a single file python simulation
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import myokit.formats.python as python

tab  = 4*' '
tab2 = 2*tab

def v(lhs, unless=None):
    """
    Translates a left-hand-side expression.

    The name is given fully qualified, unless the variable's parent comp
     is equal to the second parameter.
    """
    if isinstance(lhs, myokit.Derivative):
        var = lhs.var()
        pre = 'self' if var._parent == unless else c(var._parent)[0]
        return pre + '.d_' + var.name().lower()
    if isinstance(lhs, myokit.Name):
        var = lhs.var()
    else:
        var = lhs
    # Local variable
    out = var.uname()
    par = var.parent(myokit.Component)
    pre = 'self.' if par == unless else c(par)[0] + '.'
    return pre + out
def c(comp):
    """
    Returns the names for this comp as (Class, object)
    """
    lname = comp.name().lower()
    uname = lname[0].upper() + lname[1:]
    return ('c_' + lname, 'C' + uname)

# Create python expression writer
w = python.PythonExpressionWriter()
w.set_lhs_function(v)
e = w.eq

# Process bindings
bound_variables = myokit._prepare_bindings(model, {
    'time' : 'engine.time',
    'pace' : 'engine.pace',
    })

# Get equations
equations = model.solvable_order()
components = []
for comp in equations:
    if comp != '*remaining*':
        components.append(model.get(comp))

?>#!/usr/bin/env python
#
# Generated on <?= myokit.date() ?>
#
import math

#
# Components and variables
#
<?
# Generate class per component
for comp in components:
    names = c(comp)
    w.set_lhs_function(lambda x: v(x, comp))

    ?>
class <?= names[1] ?>:
    def __init__(self):
<?
    for var in comp.variables(deep=True):
        print(tab2 + v(var, comp) + ' = None')
        if var.is_state():
            print(tab2 + v(var.lhs(), comp) + ' = None')
    ?>
        self._constants()
        self.init()
    def _constants(self):
        """
        Sets the constant values
        """
<?
    eqs = equations[comp.qname()]
    for eq in eqs.equations(const=True):
        print(tab2 + e(eq))
    ?>    def init(self):
        """
        Resets the state variables to their initial values
        """
<?
    for var, value in zip(model.states(), model.initial_values(True)):
        if var.parent() == comp:
            print(tab2 + v(var, comp) + ' = ' + myokit.float.str(value))
    ?>    def update(self):
        """
        Re-calculates all values for the current time and state
        """
<?
    printed = False
    for eq in eqs.equations(const=False):
        printed = True
        var = eq.lhs.var()
        if var.is_bound():
            print(tab2 + v(var) + ' = ' + bound_variables[var])
        else:
            print(tab2 + e(eq))
    if not printed:
        print(tab2 + 'pass')
?>
#
# Engine component
#
class Engine:
    """
    Calculates the derivatives in the current state
    """
    def __init__(self):
        self.pace = 0.0
        self.time = 0.0
<?

# Update function, calls all others
print(tab + 'def update(self):')
for comp in components:
    print(tab2 + c(comp)[0] + '.update()')

# Now handle remaining equations
print(tab2 + '# Remaining equations')
for eq in equations['*remaining*'].equations(const=False):
    print(tab2 + e(eq))

?>
#
# Create objects, set initial values
#
def init():
    """ (Re-)Initializes the model """
<?
# Set all component instances all global variables
n = 0
for comp in components:
    names = c(comp)
    if len(names[0]) > n:
        n = len(names[0])
    print(tab + 'global ' + names[0])
print(tab + 'global ' + 'engine')
# Create component instances
for comp in components:
    names = c(comp)
    print(tab + names[0] + ' '*(n - len(names[0])) + ' = ' + names[1] + '()')
print(tab + 'engine = Engine()')
# Any remaining constants
for eq in equations['*remaining*'].equations(const=True):
    print(tab + e(eq))
?>
#
# Update function (rhs function, takes a single step)
#
def update(stepSize):
    """ Calculates all derivatives, update state, advances time """
    engine.update()
<?
m = 0
for var in model.states():
    m = max(m, len(v(var)))
for var in model.states():
    name = v(var)
    print(tab + name + ' '*(m - len(name)) + ' += stepSize * ' + v(var.eq().lhs))
?>
#
# State vector returning function
#
def state():
    """ Returns the state vector """
<?
names = [v(var) for var in model.states()]
print(tab + 'return [' + (',\n'+tab2).join(names) + ']')
?>
#
# State vector printing function
#
def print_state():
    """ Prints the current state to the screen """
<?
n = max([len(x.qname()) for x in model.states()])
n = max(n, 4)
?>
    f = "{:<<?=str(n)?>}  {:<20}  {:<20}"
    print("-"*80)
    print(f.format("Name", "State value", "Derivative"))
    f = "{: <<?=str(n)?>}  {:< 20.13e}  {:< 20.13e}"
<?
for var in model.states():
    print(tab+'print(f.format("' +var.qname()+'", '+v(var)+', '+v(var.lhs()) + '))')
?>
#
# Test step function
#
def test_step():
    """ Calculates and prints the initial derivatives """
    init()
    engine.update()
    print_state()

#
# Pacing
#
class Protocol:
    """ Holds an ordered set of ProtocolEvent objects """
    def __init__(self):
        super().__init__()
        self.head = None
    def add(self, e):
        """ Schedules an event """
        if self.head is None:
            self.head = e
            return
        if e.start < self.head.start:
            e.next = self.head
            self.head = e
            return
        f = self.head
        while (f.next is not None) and (e.start > f.next.start):
            f = f.next
        e.next = f.next
        f.next = e
    def pop(self):
        """ Returns the next event """
        e = self.head
        if self.head is not None:
            self.head = self.head.next
        return e


class ProtocolEvent:
    def __init__(self, level, start, duration, period=0, multiplier=0):
        super().__init__()
        self.level = float(level)
        self.start = float(start)
        self.duration = float(duration)
        if self.duration <= 0:
            raise Exception('Duration must be greater than zero')
        self.period = float(period)
        if self.period < 0:
            raise Exception('multiplier must be zero or greater')
        self.multiplier = int(multiplier)
        if self.multiplier < 0:
            raise Exception('Multiplier must be zero or greater')
        if self.period == 0 and self.multiplier > 0:
            raise Exception('Non-periodic event cannot occur more than once')
        self.next = None


def pacing_protocol():
    pacing = Protocol()
<?
nxt = protocol.head()
while nxt:
    x = [
        nxt.level(),
        nxt.start(),
        nxt.duration(),
        nxt.period(),
        nxt.multiplier(),
        ]
    x = ', '.join([str(x) for x in x])
    print(tab + 'pacing.add(ProtocolEvent(' + x + '))')
    nxt = nxt.next()
print(tab + 'return pacing')

vm = v(next(iter(model.states())))
?>
#
# Solver
#
def beat(stepSmall = 0.005, stepLarge = 0.01):
    """
    Simulates a single beat
    """
    tmin = 0
    tmax = 1000
    # Feedback
    outInt = int(math.ceil(tmax / 10.0))
    outPos = engine.time
    outVal = 0
    # Logging
    logInt = 1
    logPos = engine.time
    log = []
    # Stepsize
    stepSize = stepSmall
    hadPulse = False
    vInit = <?= vm ?>
    # Pacing
    pacing = pacing_protocol()
    next = pacing.head
    while next and next.start < tmin:
        next = next.next
    if next.start < tmin:
        next = None
    fire = None
    fireDown = 0
    stopTime = min(next.start, tmin + stepSize)
    print('Starting integration with step sizes ' + str(stepSmall) + ' and ' + str(stepLarge) + '.')
    while engine.time < tmax:
        update(stopTime - engine.time)
        engine.time = stopTime
        # Event over
        if (fire and engine.time >= fireDown):
            engine.pace = 0
            fire = None
        # New event
        if (next and engine.time >= next.start):
            fire = next
            next = next.next
            engine.pace = fire.level
            fireDown = fire.start + fire.duration
            if fire.period > 0:
                if fire.multiplier == 1:
                    fire.period = 0
                else:
                    if fire.multiplier > 1:
                        fire.multiplier -=1
                    fire.start += fire.period
                    pacing.add(fire)
                    next = pacing.head
        # User feedback
        if engine.time >= outPos and outVal < 100:
            print(str(outVal) + "%")
            outVal += 10
            outPos += outInt
        # Logging
        if engine.time >= logPos:
            log.append((engine.time, state()))
            logPos += logInt
        # Step size update
        if fire: # or <?= vm ?> > -70:
            if stepSize != stepSmall:
                print("Small steps")
                stepSize = stepSmall
        else:
            if stepSize != stepLarge:
                print("Big steps")
                stepSize = stepLarge
        # Set next time
        stopTime = engine.time + stepSize
        if fire and fireDown < stopTime:
            stopTime = fireDown
        if next and next.start < stopTime:
            stopTime = next.start
        if logPos < stopTime:
            stopTime = logPos
    print("100% done")
    print("t = " + str(engine.time))
    print_state()
    return log

#
# Run if loaded as main script
#
if __name__ == '__main__':
    small = 0.005
    large = 0.01
    go = True
    done = False
    while go:
        go = False
        try:
            init()
            data = beat(small, large)
            done = True
        except ArithmeticError as e:
            print('Arithmetic error occurred')
            y = 'Continue with smaller stepsize? (y/n): '
            y = input(y)
            if y.lower()[0:1] == 'y':
                small /= 2
                large /= 2
                go = True
    if done:
        print('Showing result...')
        x = []
        y = []
        for time, state in data:
            x.append(time)
            y.append(state[0])
        import matplotlib.pyplot as py
        plot = py.plot(x, y)
        py.show()



#!/usr/bin/env python3
#
# Class for testing the time-series pacing system.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit

from myokit.tests import DIR_TEST

# Location of C template
SOURCE_FILE = 'ansic_time_series_pacing.c'


class AnsicTimeSeriesPacing(myokit.CModule):
    """
    Class for testing the time-series pacing system.
    """
    _index = 0

    def __init__(self, protocol):
        super().__init__()

        # Unique id
        AnsicTimeSeriesPacing._index += 1
        module_name = \
            'myokit_ansic_fpacing_' + str(AnsicTimeSeriesPacing._index)

        # Arguments
        fname = os.path.join(DIR_TEST, SOURCE_FILE)

        # Create back-end
        args = {'module_name': module_name}
        libs = []
        libd = []
        incd = [DIR_TEST, myokit.DIR_CFUNC]
        self._sys = self._compile(module_name, fname, args, libs, libd, incd)

        # Initialize
        self._sys.init(protocol)

    def pace(self, time):
        return self._sys.pace(time)

    def __del__(self):
        # Free the memory used by the pacing system
        try:
            self._sys.clean()
        except AttributeError:
            pass



#
# Provides CellML support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import CellMLExporter, CellML1Exporter, CellML2Exporter
from ._ewriter import CellMLExpressionWriter
from ._importer import CellMLImporter, CellMLImporterError  # noqa


# Namespaces
NS_BQBIOL = 'http://biomodels.net/biology-qualifiers/'
NS_CELLML_1_0 = 'http://www.cellml.org/cellml/1.0#'
NS_CELLML_1_1 = 'http://www.cellml.org/cellml/1.1#'
NS_CELLML_2_0 = 'http://www.cellml.org/cellml/2.0#'
NS_CMETA = 'http://www.cellml.org/metadata/1.0#'
NS_MATHML = 'http://www.w3.org/1998/Math/MathML'
NS_OXMETA = 'https://chaste.comlab.ox.ac.uk/cellml/ns/oxford-metadata#'
NS_RDF = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
NS_TMP_DOC = 'http://cellml.org/tmp-documentation'


# Importers
_importers = {
    'cellml': CellMLImporter,
}


def importers():
    """
    Returns a dict of all importers available in this module.
    """
    return dict(_importers)


# Exporters
_exporters = {
    'cellml': CellMLExporter,
    'cellml1': CellML1Exporter,
    'cellml2': CellML2Exporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'cellml': CellMLExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)




#
# Source code editor for Myokit.
#
# This code is based in part on examples provided by the PyQt project.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.gui import Qt, QtCore, QtGui, QtWidgets


# GUI components
# Constants
SPACE = ' '
TABS = 4
INDENT = SPACE * TABS
BRACKETS = {
    '(': ')',
    ')': '(',
    '[': ']',
    ']': '['
}
BRACKETS_CLOSE = (')', ']')
FONT = myokit.gui.qtMonospaceFont()
FONT.setPointSize(11)

# Component and model headers
STYLE_HEADER = QtGui.QTextCharFormat()

# Comments
STYLE_COMMENT = QtGui.QTextCharFormat()

# Model annotations (including meta, labels, and units)
STYLE_ANNOT_KEY = QtGui.QTextCharFormat()
STYLE_ANNOT_VAL = QtGui.QTextCharFormat()

# Language keywords
STYLE_KEYWORD_1 = QtGui.QTextCharFormat()
STYLE_KEYWORD_2 = QtGui.QTextCharFormat()

# Literals: Numbers in model/protocol, Also booleans and strings in script
STYLE_LITERAL = QtGui.QTextCharFormat()
STYLE_INLINE_UNIT = QtGui.QTextCharFormat()

# Matching brackets are highlighted
COLOR_BRACKET = QtGui.QColor(240, 100, 0)

# Selected line is highlighted
COLOR_SELECTED_LINE = QtGui.QColor(238, 238, 238)


def _check_for_dark_mode(palette):
    """
    Checks the default editor background color, and adjusts the color scheme
    if it looks like dark-mode is enabled.
    """
    c = palette.base().color()
    c = (c.blueF() + c.greenF() + c.redF()) / 3
    dark = c < 0.5

    # Don't mess with these directly: Use the SVG in myokit-docs
    if not dark:
        STYLE_HEADER.setForeground(QtGui.QColor(0, 31, 231))
        STYLE_COMMENT.setForeground(QtGui.QColor(103, 161, 107))
        STYLE_ANNOT_KEY.setForeground(QtGui.QColor(0, 31, 231))
        STYLE_ANNOT_VAL.setForeground(QtGui.QColor(57, 115, 214))
        STYLE_KEYWORD_1.setForeground(QtGui.QColor(0, 128, 0))
        STYLE_KEYWORD_1.setFontWeight(QtGui.QFont.Weight.Bold)
        STYLE_KEYWORD_2.setForeground(QtGui.QColor(0, 128, 128))
        STYLE_LITERAL.setForeground(QtGui.QColor(255, 20, 215))
        STYLE_INLINE_UNIT.setForeground(QtGui.QColor(128, 0, 128))
    else:
        STYLE_HEADER.setForeground(QtGui.QColor(98, 178, 255))
        STYLE_COMMENT.setForeground(QtGui.QColor(153, 153, 153))
        STYLE_ANNOT_KEY.setForeground(QtGui.QColor(179, 179, 179))
        STYLE_ANNOT_VAL.setForeground(QtGui.QColor(171, 177, 205))
        STYLE_KEYWORD_1.setForeground(QtGui.QColor(10, 195, 87))
        STYLE_KEYWORD_1.setFontWeight(QtGui.QFont.Weight.Bold)
        STYLE_KEYWORD_2.setForeground(QtGui.QColor(10, 195, 87))
        STYLE_LITERAL.setForeground(QtGui.QColor(255, 223, 12))
        STYLE_INLINE_UNIT.setForeground(QtGui.QColor(168, 152, 33))

        global COLOR_SELECTED_LINE
        COLOR_SELECTED_LINE = QtGui.QColor(70, 70, 70)


# Classes & methods
class Editor(QtWidgets.QPlainTextEdit):
    """
    Source code editor used in Myokit.

    Provides the signal ``find_action(str)`` which is fired everything a find
    action occurred with a description that can be used in an application's
    status bar.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        # Current style
        self._palette = QtGui.QGuiApplication.palette()
        _check_for_dark_mode(self._palette)

        # Apply default settings
        self._default_settings()

        # Add line number area
        self._line_number_area = LineNumberArea(self)
        self._line_number_area.update_width(0)

        # Add current line highlighting and bracket matching
        self.cursorPositionChanged.connect(self.cursor_changed)
        self.cursor_changed()

        # Line position
        try:
            # https://doc.qt.io/qt-5/qfontmetrics.html#horizontalAdvance
            # Qt 5.5.11 and onwards
            self._line_offset = self.fontMetrics().horizontalAdvance(' ' * 79)
        except AttributeError:
            self._line_offset = self.fontMetrics().width(' ' * 79)

        # Number of blocks in page up/down
        self._blocks_per_page = 1

        # Last position in line, used for smart up/down buttons
        self._last_column = None
        self.textChanged.connect(self._text_has_changed)

    def cursor_changed(self):
        """ Slot: Called when the cursor position is changed """
        # Highlight current line
        extra_selections = []
        selection = QtWidgets.QTextEdit.ExtraSelection()
        selection.format.setBackground(COLOR_SELECTED_LINE)
        selection.format.setProperty(
            QtGui.QTextFormat.Property.FullWidthSelection, True)
        selection.cursor = self.textCursor()
        selection.cursor.clearSelection()
        extra_selections.append(selection)

        # Bracket matching
        cursor = self.textCursor()
        if not cursor.hasSelection():
            # Test if in front of or behind an opening or closing bracket
            pos = cursor.position()
            bracket = None
            if not cursor.atEnd():
                cursor.setPosition(
                    pos + 1, QtGui.QTextCursor.MoveMode.KeepAnchor)
                text = cursor.selectedText()
                if text in BRACKETS:
                    bracket = cursor
            elif bracket is None and not cursor.atStart():
                cursor.setPosition(pos - 1)
                cursor.setPosition(pos, QtGui.QTextCursor.MoveMode.KeepAnchor)
                text = cursor.selectedText()
                if text in BRACKETS:
                    bracket = cursor

            if bracket:
                # Find matching partner
                doc = self.document()
                depth = 1
                start = bracket.position()
                while depth > 0:
                    if text in BRACKETS_CLOSE:
                        other = doc.find(
                            text, start - 1,
                            QtGui.QTextDocument.FindFlag.FindBackward)
                        match = doc.find(
                            BRACKETS[text], start - 1,
                            QtGui.QTextDocument.FindFlag.FindBackward)
                    else:
                        other = doc.find(text, start)
                        match = doc.find(BRACKETS[text], start)
                    if match.isNull():
                        break
                    if other.isNull():
                        depth -= 1
                        start = match.position()
                    elif text in BRACKETS_CLOSE:
                        if other.position() < match.position():
                            depth -= 1
                            start = match.position()
                        else:
                            depth += 1
                            start = other.position()
                    else:
                        if match.position() < other.position():
                            depth -= 1
                            start = match.position()
                        else:
                            depth += 1
                            start = other.position()
                if depth == 0:
                    # Apply formatting
                    selection = QtWidgets.QTextEdit.ExtraSelection()
                    selection.cursor = bracket
                    selection.format.setBackground(self._palette.mid())
                    selection.format.setForeground(COLOR_BRACKET)
                    extra_selections.append(selection)
                    selection = QtWidgets.QTextEdit.ExtraSelection()
                    selection.cursor = match
                    selection.format.setBackground(self._palette.mid())
                    selection.format.setForeground(COLOR_BRACKET)
                    extra_selections.append(selection)

        if extra_selections:
            self.setExtraSelections(extra_selections)

    def cursor_position(self):
        """
        Returns a tuple ``(line, char)`` with the current cursor position. If
        a selection is made only the left position is used.

        Line and char counts both start at zero.
        """
        cursor = self.textCursor()
        line = cursor.blockNumber()
        char = cursor.selectionStart() - cursor.block().position()
        return (line, char)

    def _default_settings(self):
        """ Applies this editor's default settings. """
        # Set font
        self.setFont(FONT)
        # Set frame
        self.setFrameStyle(
            QtWidgets.QFrame.Shape.WinPanel | QtWidgets.QFrame.Shadow.Sunken)
        # Disable wrapping
        self.setLineWrapMode(QtWidgets.QPlainTextEdit.LineWrapMode.NoWrap)
        # Set tab width (if ever seen) to 4 spaces
        try:
            # https://doc.qt.io/qt-5/qtextedit-obsolete.html
            # https://doc.qt.io/qt-5/qfontmetrics.html#horizontalAdvance
            # Qt 5.10/5.11 and onwards
            ts = self.fontMetrics().horizontalAdvance(' ' * 4)
            self.setTabStopDistance(ts)
        except AttributeError:
            ts = self.fontMetrics().width(' ' * 4)
            self.setTabStopWidth(ts)

    def get_text(self):
        """ Returns the text in this editor. """
        return self.toPlainText()

    def jump_to(self, line, char):
        """ Jumps to the given line and row (with indices starting at 0). """
        block = self.document().findBlockByNumber(line)
        cursor = self.textCursor()
        cursor.setPosition(block.position() + char)
        self.setTextCursor(cursor)
        self.centerCursor()

    def keyPressEvent(self, event):
        """ Qt event: A key was pressed. """
        K = Qt.Key
        KM = Qt.KeyboardModifier
        MM = QtGui.QTextCursor.MoveMode
        MO = QtGui.QTextCursor.MoveOperation

        # Get key and modifiers
        key = event.key()
        mod = event.modifiers()
        # Possible modifiers:
        #  NoModifier
        #  ShiftModifier, ControlModifier, AltModifiier
        #  MetaModifier (i.e. super key)
        #  KeyPadModifier (button is part of keypad)
        #  GroupSwitchModifier (x11 thing)

        # Ignore the keypad modifier, we don't care!
        if mod & KM.KeypadModifier:
            mod = mod ^ KM.KeypadModifier   # xor!

        # Actions per key/modifier combination
        if key == K.Key_Tab and mod == KM.NoModifier:
            # Indent
            cursor = self.textCursor()
            start, end = cursor.selectionStart(), cursor.selectionEnd()
            if cursor.hasSelection():
                # Add single tab to all lines in selection
                cursor.beginEditBlock()     # Undo grouping
                doc = self.document()
                b = doc.findBlock(start)
                e = doc.findBlock(end).next()
                while b != e:
                    cursor.setPosition(b.position())
                    cursor.insertText(TABS * SPACE)
                    b = b.next()
                cursor.endEditBlock()
            else:
                # Insert spaces until next tab stop
                pos = cursor.positionInBlock()
                cursor.insertText((TABS - pos % TABS) * SPACE)

        elif key == K.Key_Backtab and mod == KM.ShiftModifier:
            # Dedent all lines in selection (or single line if no selection)
            '''
            cursor = self.textCursor()
            start, end = cursor.selectionStart(), cursor.selectionEnd()
            cursor.beginEditBlock() # Undo grouping
            doc = self.document()
            # Get blocks in selection
            blocks = []
            b = doc.findBlock(start)
            while b.isValid() and b.position() <= end:
                blocks.append(b)
                b = b.next()
            # Dedent
            for b in blocks:
                t = b.text()
                p1 = b.position()
                p2 = p1 + min(4, len(t) - len(t.lstrip()))
                c = self.textCursor()
                c.setPosition(p1)
                c.setPosition(p2, MM.KeepAnchor)
                c.removeSelectedText()
            cursor.endEditBlock()
            '''
            # This silly method is required because of a bug in qt5 (and 6?)
            cursor = self.textCursor()
            start, end = cursor.selectionStart(), cursor.selectionEnd()
            first = self.document().findBlock(start)
            q = 0
            new_text = []
            new_start, new_end = start, end
            b = QtGui.QTextBlock(first)
            while b.isValid() and b.position() <= end:
                t = b.text()
                p = min(4, len(t) - len(t.lstrip()))
                new_text.append(t[p:])
                if b == first:
                    new_start -= p
                new_end -= p
                q += p
                b = b.next()
            last = b.previous()
            new_start = max(new_start, first.position())
            new_end = max(new_end, new_start)
            if q > 0:
                # Cut text, replace with new
                cursor.beginEditBlock()
                cursor.setPosition(first.position())
                cursor.setPosition(
                    last.position() + last.length() - 1, MM.KeepAnchor)
                cursor.removeSelectedText()
                cursor.insertText('\n'.join(new_text))
                cursor.endEditBlock()
                # Set new cursor
                cursor.setPosition(new_start)
                cursor.setPosition(new_end, MM.KeepAnchor)
                self.setTextCursor(cursor)

        elif key == K.Key_Enter or key == K.Key_Return:
            # Enter/Return with modifier is overruled here to mean nothing
            # This is very important as the default for shift-enter is to
            # start a new line within the same block (this can't happen with
            # copy-pasting, so it's safe to just catch it here).
            if mod == KM.NoModifier:
                # "Smart" enter:
                #   - If selection, selection is deleted
                #   - Else, autoindenting is performed
                cursor = self.textCursor()
                cursor.beginEditBlock()
                if cursor.hasSelection():
                    # Replace selection with newline,
                    cursor.removeSelectedText()
                    cursor.insertBlock()
                else:
                    # Insert new line with correct indenting
                    b = self.document().findBlock(cursor.position())
                    t = b.text()
                    i = t[:len(t) - len(t.lstrip())]
                    i = i[:cursor.positionInBlock()]
                    cursor.insertBlock()
                    cursor.insertText(i)
                cursor.endEditBlock()
                # Scroll if necessary
                self.ensureCursorVisible()

        elif key == K.Key_Home and (
                mod == KM.NoModifier or mod == KM.ShiftModifier):
            # Plain home button: move to start of line
            # If Control is used: Jump to start of document
            # Ordinary home button: Jump to first column or first
            # non-whitespace character
            cursor = self.textCursor()
            block = cursor.block()
            cp = cursor.position()
            bp = block.position()
            if cp != bp:
                # Jump to first column
                newpos = bp
                # Smart up/down:
                self._last_column = 0
            else:
                # Already at first column: Jump to first non-whitespace or
                # end of line if all whitespace
                t = block.text()
                indent = len(t) - len(t.lstrip())
                newpos = bp + indent
                # Smart up/down:
                self._last_column = indent
            # If Shift is used: only move position (keep anchor, i.e. select)
            cursor.setPosition(
                newpos,
                MM.KeepAnchor if mod == KM.ShiftModifier else MM.MoveAnchor)
            self.setTextCursor(cursor)

        elif key == K.Key_Home and (
                mod == KM.ControlModifier
                or mod == KM.ControlModifier & KM.ShiftModifier):
            # Move to start of document
            # If Shift is used: only move position (keep anchor, i.e. select)
            cursor = self.textCursor()
            cursor.setPosition(
                0, MM.KeepAnchor if mod == KM.ShiftModifier else MM.MoveAnchor)
            self.setTextCursor(cursor)

        elif key in (K.Key_Up, K.Key_Down) and mod == KM.AltModifier:
            # Move selected lines up or down
            # Get current selection
            doc = self.document()
            cursor = self.textCursor()
            start, end = cursor.selectionStart(), cursor.selectionEnd()
            block1 = doc.findBlock(start)
            if start == end:
                block2 = block1
            else:
                block2 = doc.findBlock(end)
                # Whole line selection? Then move end back 1 position
                if end == block2.position():
                    end -= 1
                    block2 = block2.previous()  # always valid
            block2 = block1 if start == end else doc.findBlock(end)
            # Check if we can move
            if key == K.Key_Up:
                if not block1.previous().isValid():
                    return
            elif not block2.next().isValid():
                return
            # Select full line(s)
            b1pos = block1.position()
            cursor.beginEditBlock()
            cursor.setPosition(b1pos)
            cursor.setPosition(end, MM.KeepAnchor)
            cursor.movePosition(MO.EndOfLine, MM.KeepAnchor)
            line = cursor.selectedText()
            size = cursor.selectionEnd() - cursor.selectionStart()
            cursor.removeSelectedText()
            if key == K.Key_Up:
                cursor.deletePreviousChar()
                cursor.movePosition(MO.StartOfLine)
                cursor.insertText(line + '\n')
                cursor.movePosition(MO.Left)
            else:
                cursor.deleteChar()
                cursor.movePosition(MO.EndOfLine)
                cursor.insertText('\n' + line)
            cursor.endEditBlock()
            # Cursor is at the end of the moved lines.
            # Set moved lines as selection
            cursor.movePosition(MO.Left, MM.KeepAnchor, size)
            self.setTextCursor(cursor)

        elif key in (K.Key_Up, K.Key_Down, K.Key_PageUp, K.Key_PageDown) \
                and (mod == KM.NoModifier or mod == KM.ShiftModifier):
            # Move cursor up/down
            # Maintain the column position, even when the current row doesn't
            # have as many characters. Reset this behavior as soon as a
            # left/right home/end action is made or whenever the text is
            # changed.
            # Set up operation
            anchor = (
                MM.KeepAnchor if mod == KM.ShiftModifier else MM.MoveAnchor)
            operation = (MO.PreviousBlock if key in (K.Key_Up, K.Key_PageUp)
                         else MO.NextBlock)
            n = 1 if key in (K.Key_Up, K.Key_Down) else (
                self._blocks_per_page - 3)

            # Move
            cursor = self.textCursor()
            if self._last_column is None:
                # Update "smart" column
                self._last_column = cursor.positionInBlock()
            if cursor.movePosition(operation, anchor, n):
                column = min(cursor.block().length() - 1, self._last_column)
                cursor.setPosition(cursor.position() + column, anchor)
            else:
                # Up/Down beyond document start/end? Move cursor to document
                # start/end and update last column
                if operation == MO.NextBlock:
                    cursor.movePosition(MO.EndOfBlock, anchor)
                else:
                    cursor.movePosition(MO.StartOfBlock, anchor)
                self._last_column = cursor.positionInBlock()
            self.setTextCursor(cursor)

        elif key in (K.Key_Left, K.Key_Right, K.Key_End) and not (
                mod & KM.AltModifier):
            # Allow all modifiers except alt
            # Reset smart up/down behavior
            self._last_column = None
            # Pass to parent class
            super().keyPressEvent(event)

        elif key == K.Key_Insert and mod == KM.NoModifier:
            # Insert/replace
            self.setOverwriteMode(not self.overwriteMode())

        else:
            # Default keyboard shortcuts / functions:
            # Backspace             OK
            # Delete                OK
            # Control+C             OK
            # Control+V             OK
            # Control+X             OK
            # Control+Insert        OK
            # Shift+Insert          OK
            # Shift+Delete          OK
            # Control+Z             OK
            # Control+Y             OK
            # LeftArrow             Overwritten (maintained)
            # RightArrow            Overwritten (maintained)
            # UpArrow               Overwritten (maintained)
            # DownArrow             Overwritten (maintained)
            # Control+RightArrow    OK (Jump to next word)
            # Control+LeftArrow     OK (Jump to previous word)
            # Control+UpArrow       Removed
            # Control+Down Arrow    Removed
            # PageUp                Overwritten (maintained)
            # PageDown              Overwritten (maintained)
            # Home                  Overwritten (maintained)
            # End                   Overwritten (maintained)
            # Control+Home          Overwritten (maintained)
            # Control+End           Overwritten (maintained)
            # Alt+Wheel             OK (Horizontal scrolling)
            # Control+Wheel         OK (Fast scrolling)
            # Control+K             Removed
            # Not listed, but very important:
            # Shift-Enter           Starts new line within the same block!
            #                       Definitely removed
            # Ctrl-i                Undocumented, but inserts tab...
            ctrl_ignore = (K.Key_K, K.Key_I)
            if mod == KM.ControlModifier and key in ctrl_ignore:
                # Control-K: ignore
                pass
            elif key == K.Key_Up or key == K.Key_Down:
                # Up/down with modifiers: ignore
                pass
            else:
                # Let parent class handle it
                super().keyPressEvent(event)

    def _line_number_area_width(self):
        """ Returns the required width for the number area. """
        text = str(max(1, self.blockCount()))
        try:
            # https://doc.qt.io/qt-5/qfontmetrics.html#horizontalAdvance
            # Qt 5.5.11 and onwards
            return 8 + self.fontMetrics().horizontalAdvance(text)
        except AttributeError:
            return 8 + self.fontMetrics().width(text)

    def _line_number_area_paint(self, area, event):
        """ Repaints the line number area. """
        # Area to repaint
        rect = event.rect()
        etop = rect.top()
        ebot = rect.bottom()

        # Font metrics
        metrics = self.fontMetrics()
        height = metrics.height()
        width = area.width()

        # Create painter, set font color
        painter = QtGui.QPainter(area)
        painter.fillRect(rect, self._palette.button())
        painter.setPen(self._palette.buttonText().color())

        # Get top and bottom of first visible block
        block = self.firstVisibleBlock()
        geom = self.blockBoundingGeometry(block)
        btop = int(geom.translated(self.contentOffset()).top())
        bbot = int(btop + geom.height())

        # Iterate over visible blocks
        count = block.blockNumber()
        while block.isValid() and btop <= ebot:
            count += 1
            if block.isVisible() and bbot >= etop:
                painter.drawText(0, btop, width - 4, height,
                                 Qt.AlignmentFlag.AlignRight, str(count))
            block = block.next()
            btop = bbot
            bbot += int(self.blockBoundingRect(block).height())

    def paintEvent(self, e):
        """ Paints this editor. """

        # Paint the editor
        super().paintEvent(e)

        # Paint a line between the editor and the line number area
        x = int(
            self.contentOffset().x()
            + self.document().documentMargin()
            + self._line_offset
        )
        p = QtGui.QPainter(self.viewport())
        p.setPen(QtGui.QPen(QtGui.QColor('#ddd')))
        rect = e.rect()
        p.drawLine(x, rect.top(), x, rect.bottom())

    def replace(self, text):
        """
        Replaces the current text with the given text, in a single operation
        that does not reset undo/redo.
        """
        self.selectAll()
        cursor = self.textCursor()
        cursor.beginEditBlock()
        cursor.removeSelectedText()
        self.appendPlainText(str(text))
        cursor.endEditBlock()

    def resizeEvent(self, event):
        """ Qt event: Editor is resized. """
        super().resizeEvent(event)
        # Update line number area
        rect = self.contentsRect()
        self._line_number_area.setGeometry(
            rect.left(), rect.top(),
            self._line_number_area_width(), rect.height())
        # Set number of "blocks" per page
        font = self.fontMetrics()
        self._blocks_per_page = int(rect.height() / font.height())

    def set_cursor(self, pos):
        """
        Changes the current cursor to the given position and scrolls so that
        its visible.
        """
        cursor = self.textCursor()
        cursor.setPosition(pos)
        self.setTextCursor(cursor)
        self.centerCursor()

    def set_text(self, text):
        """ Replaces the text in this editor. """
        if text:
            self.setPlainText(str(text))
        else:
            # Bizarre workaround for bug:
            #   https://bugreports.qt.io/browse/QTBUG-42318
            self.selectAll()
            cursor = self.textCursor()
            cursor.removeSelectedText()
            doc = self.document()
            doc.clearUndoRedoStacks()
            doc.setModified(False)

    def _text_has_changed(self):
        """
        Called whenever the text has changed, resets the smart up/down
        behavior.
        """
        self._last_column = None

    def toggle_comment(self):
        """ Comments or uncomments the selected lines """
        # Comment or uncomment selected lines
        cursor = self.textCursor()
        start, end = cursor.selectionStart(), cursor.selectionEnd()
        doc = self.document()
        first, last = doc.findBlock(start), doc.findBlock(end)
        # Determine minimum indent and adding or removing
        block = first
        blocks = [first]
        while block != last:
            block = block.next()
            blocks.append(block)
        lines = [block.text() for block in blocks]
        indent = [len(t) - len(t.lstrip()) for t in lines if len(t) > 0]
        indent = min(indent) if indent else 0
        remove = True
        for line in lines:
            if line[indent:indent + 1] != '#':
                remove = False
                break
        cursor.beginEditBlock()
        if remove:
            for block in blocks:
                p = block.position() + indent
                cursor.setPosition(p)
                cursor.setPosition(
                    p + 1, QtGui.QTextCursor.MoveMode.KeepAnchor)
                cursor.removeSelectedText()
        else:

            for block in blocks:
                p = block.position()
                n = len(block.text())
                if len(block.text()) < indent:
                    cursor.setPosition(p)
                    cursor.setPosition(
                        p + n, QtGui.QTextCursor.MoveMode.KeepAnchor)
                    cursor.removeSelectedText()
                    cursor.insertText(' ' * indent + '#')
                else:
                    cursor.setPosition(p + indent)
                    cursor.insertText('#')
        cursor.endEditBlock()

    def trim_trailing_whitespace(self):
        """ Trims all trailing whitespace from this document. """
        block = self.document().begin()
        cursor = self.textCursor()
        cursor.beginEditBlock()     # Undo grouping
        while block.isValid():
            t = block.text()
            a = len(t)
            b = len(t.rstrip())
            if a > b:
                cursor.setPosition(block.position() + b)
                cursor.setPosition(block.position() + a,
                                   QtGui.QTextCursor.MoveMode.KeepAnchor)
                cursor.removeSelectedText()
            block = block.next()
        cursor.endEditBlock()


class LineNumberArea(QtWidgets.QWidget):
    """
    Line number area widget for the editor. All real actions are delegated to
    the text area class.

    The line number is drawn in the left margin of the :class:`Editor` widget,
    the space to do so is created by setting the editor's viewport margins.
    """

    def __init__(self, editor):
        super().__init__(editor)
        self._editor = editor
        self._editor.blockCountChanged.connect(self.update_width)
        self._editor.updateRequest.connect(self.update_contents)

    def paintEvent(self, event):
        """ Qt event: Paint this area. """
        self._editor._line_number_area_paint(self, event)

    def sizeHint(self):
        """ Qt event: Suggest a size for this area. """
        return QtCore.QSize(self._editor._line_number_area_width(), 0)

    def update_contents(self, rect, scroll):
        """
        Slot: Invoked when the text editor view has changed and the line
        numbers need to be redrawn.
        """
        if scroll:
            # Scroll
            self.scroll(0, scroll)
        else:
            self.update()

    def update_width(self, count):
        """
        Slot: Invoked when the number of lines in the text area changed, which
        might change the size of the number area.
        """
        # Update the editor margins, so that the line number area can be
        # painted in the margins.
        self._editor.setViewportMargins(
            2 + self._editor._line_number_area_width(), 0, 0, 0)


class FindReplaceWidget(QtWidgets.QWidget):
    """
    Find/replace widget for :class:`Editor`.
    """
    # Signal: Find action happened, update with text
    # Attributes: (description)
    find_action = QtCore.Signal(str)

    def __init__(self, parent, editor):
        super().__init__(parent)
        self._editor = editor

        # Create widgets
        self._replace_all_button = QtWidgets.QPushButton('Replace all')
        self._replace_all_button.clicked.connect(self.action_replace_all)
        self._replace_button = QtWidgets.QPushButton('Replace')
        self._replace_button.clicked.connect(self.action_replace)
        self._find_button = QtWidgets.QPushButton('Find')
        self._find_button.clicked.connect(self.action_find)
        self._search_label = QtWidgets.QLabel('Search for')
        self._search_field = QtWidgets.QLineEdit()
        self._replace_label = QtWidgets.QLabel('Replace with')
        self._replace_field = QtWidgets.QLineEdit()
        self._case_check = QtWidgets.QCheckBox('Case sensitive')
        self._whole_check = QtWidgets.QCheckBox('Match whole word only')

        # Create layout
        text_layout = QtWidgets.QGridLayout()
        text_layout.addWidget(self._search_label, 0, 0)
        text_layout.addWidget(self._search_field, 0, 1)
        text_layout.addWidget(self._replace_label, 1, 0)
        text_layout.addWidget(self._replace_field, 1, 1)
        check_layout = QtWidgets.QBoxLayout(
            QtWidgets.QBoxLayout.Direction.TopToBottom)
        check_layout.addWidget(self._case_check)
        check_layout.addWidget(self._whole_check)
        button_layout = QtWidgets.QGridLayout()
        button_layout.addWidget(self._replace_all_button, 0, 1)
        button_layout.addWidget(self._replace_button, 0, 2)
        button_layout.addWidget(self._find_button, 0, 3)

        layout = QtWidgets.QBoxLayout(
            QtWidgets.QBoxLayout.Direction.TopToBottom)
        layout.addLayout(text_layout)
        layout.addLayout(check_layout)
        layout.addLayout(button_layout)
        layout.addStretch(1)
        self.setLayout(layout)

        # Accept keyboard focus on search and replace fields
        self._search_field.setEnabled(True)
        self._replace_field.setEnabled(True)

    def action_find(self):
        """ Qt slot: Find (next) item. """
        query = self._search_field.text()
        if query == '':
            self.find_action.emit('No query set')
            return
        flags = QtGui.QTextDocument.FindFlag(0)
        if self._case_check.isChecked():
            flags |= QtGui.QTextDocument.FindFlag.FindCaseSensitively
        if self._whole_check.isChecked():
            flags |= QtGui.QTextDocument.FindFlag.FindWholeWords
        if flags:
            found = self._editor.find(query, flags)
        else:
            found = self._editor.find(query)
        if found is False:
            # Not found? Try from top of document
            previous_cursor = self._editor.textCursor()
            previous_scroll = self._editor.verticalScrollBar().value()
            cursor = self._editor.textCursor()
            cursor.setPosition(0)
            self._editor.setTextCursor(cursor)
            if flags:
                found = self._editor.find(query, flags)
            else:
                found = self._editor.find(query)
            if found is False:
                self._editor.setTextCursor(previous_cursor)
                self._editor.verticalScrollBar().setValue(previous_scroll)
                self.find_action.emit('Query not found.')
                return
        cursor = self._editor.textCursor()
        line = 1 + cursor.blockNumber()
        char = cursor.selectionStart() - cursor.block().position()
        self.find_action.emit(
            'Match found on line ' + str(line) + ' char ' + str(char) + '.')

    def action_replace(self):
        """ Qt slot: Replace found item with replacement. """
        query = self._search_field.text()
        replacement = self._replace_field.text()
        if query == '':
            self.find_action.emit('No query set')
            return
        cursor = self._editor.textCursor()
        a, b = cursor.selectedText(), query
        if not self._case_check.isChecked():
            a, b = a.lower(), b.lower()
        if a == b:
            cursor.insertText(replacement)
        self.action_find()

    def action_replace_all(self):
        """ Qt slot: Replace all found items with replacement """
        query = self._search_field.text()
        replacement = self._replace_field.text()
        if query == '':
            self.find_action.emit('No query set')
            return
        flags = QtGui.QTextDocument.FindFlag(0)
        if self._case_check.isChecked():
            flags |= QtGui.QTextDocument.FindFlag.FindCaseSensitively
        if self._whole_check.isChecked():
            flags |= QtGui.QTextDocument.FindFlag.FindWholeWords
        n = 0
        found = True
        scrollpos = self._editor.verticalScrollBar().value()
        grouping = self._editor.textCursor()
        grouping.beginEditBlock()
        continue_from_top = True
        while found:
            if flags:
                found = self._editor.find(query, flags)
            else:
                found = self._editor.find(query)
            if not found and continue_from_top:
                # Not found? Try from top of document
                cursor = self._editor.textCursor()
                cursor.setPosition(0)
                self._editor.setTextCursor(cursor)
                if flags:
                    found = self._editor.find(query, flags)
                else:
                    found = self._editor.find(query)
                # Don't keep going round and round
                # (This can happen if you replace something with itself, or
                # with a different case version of itself in a case-insensitive
                # search).
                continue_from_top = False
            if found:
                cursor = self._editor.textCursor()
                cursor.insertText(replacement)
                n += 1
        grouping.endEditBlock()
        self._editor.setTextCursor(grouping)
        self._editor.verticalScrollBar().setValue(scrollpos)
        self.find_action.emit('Replaced ' + str(n) + ' occurrences.')

    def activate(self):
        """ Updates the contents of the search field and gives it focus. """
        cursor = self._editor.textCursor()
        if cursor.hasSelection():
            self._search_field.setText(cursor.selectedText())
        self._search_field.selectAll()
        self._search_field.setFocus()

    def keyPressEvent(self, event):
        """ Qt event: A key-press reaches the widget. """
        key = event.key()
        if key == Qt.Key.Key_Enter or key == Qt.Key.Key_Return:
            self.action_find()
        else:
            super().keyPressEvent(event)

    def load_config(self, config, section):
        """
        Loads this search's configuration using the given :class:`ConfigParser`
        ``config``. Loads all settings from the section ``section``.
        """
        if config.has_section(section):
            # Find options: case sensitive / whole word
            if config.has_option(section, 'case_sensitive'):
                self._case_check.setChecked(
                    config.getboolean(section, 'case_sensitive'))
            if config.has_option(section, 'whole_word'):
                self._whole_check.setChecked(
                    config.getboolean(section, 'whole_word'))

    def save_config(self, config, section):
        """
        Saves this search's configuration using the given :class:`ConfigParser`
        ``config``. Stores all settings in the section ``section``.
        """
        config.add_section(section)
        # Find options: case sensitive / whole word
        config.set(section, 'case_sensitive', self._case_check.isChecked())
        config.set(section, 'whole_word', self._whole_check.isChecked())


class ModelHighlighter(QtGui.QSyntaxHighlighter):
    """
    Syntax highlighter for ``mmt`` model definitions.
    """
    KEYWORD_1 = ['use', 'as']
    KEYWORD_2 = ['and', 'or', 'not']
    ANNOT_KEYS = ['in', 'bind', 'label']

    def __init__(self, document):
        super().__init__(document)

        # Expressions used to find strings & comments
        R = QtCore.QRegularExpression
        self._string = R(r'"""')

        # Headers
        name = r'[a-zA-Z]+[a-zA-Z0-9_]*'
        self._rule_head = R(r'^\s*(\[{1,2}' + name + '\]{1,2})')

        # Simple rules
        self._rules = []

        # Numbers
        pattern = R(r'\b[+-]?[0-9]*\.?[0-9]+([eE][+-]?[0-9]+)?\b')
        self._rules.append((pattern, STYLE_LITERAL))
        unit = r'\[([a-zA-Z0-9/^-]|\*)+\]'
        self._rules.append((R(unit), STYLE_INLINE_UNIT))

        # Keywords
        for keyword in self.KEYWORD_1:
            self._rules.append((R(r'\b' + keyword + r'\b'), STYLE_KEYWORD_1))
        for keyword in self.KEYWORD_2:
            self._rules.append((R(r'\b' + keyword + r'\b'), STYLE_KEYWORD_2))

        # Meta-data coloring
        self._rules_labels = [
            R(r'(\s*)(bind)\s+(' + name + ')'),
            R(r'(\s*)(label)\s+(' + name + ')'),
        ]
        self._rule_meta = R(r'^\s*(' + name + r':)(\s*)(.+)')
        self._rule_var_unit = R(r'^(\s*)(in)(\s*)(' + unit + ')')

        # Comment
        self._comment = R(r'#')

    def _highlight_ok(self, strings, start, length):
        """ Checks if the string ``start`` to ``length`` needs formatted. """
        for lo, hi in strings:
            if lo <= start < hi or lo <= start + length < hi:
                return False
        return True

    def highlightBlock(self, text):
        """ Qt: Called whenever a block should be highlighted. """

        # To avoid formatting within strings each is stored as a (start, end).
        strings = []
        # If the start has been handled, set the offset.
        offset = 0
        # If the end has been handled, chop it off the string

        # Multi-line strings are done first, because they overrule a lot of
        # things and we can skip formatting if we're inside one.
        # Block states: 0=No string, 1=A """ string
        self.setCurrentBlockState(0)

        # Continuing a multi-line string?
        if self.previousBlockState() == 1:
            # Search for string stop
            ms = self._string.match(text)
            if ms.hasMatch():
                # Terminate the multi-line string
                offset = ms.capturedEnd(0)
                self.setFormat(0, offset, STYLE_ANNOT_VAL)
            else:
                # Whole line in the string
                self.setCurrentBlockState(1)
                self.setFormat(0, len(text), STYLE_ANNOT_VAL)
                return
        else:
            # Search for string start
            ms = self._string.match(text)
            if ms.hasMatch():
                # Potential start, but check that it's not commented out
                start = ms.capturedStart()
                mc = self._comment.match(text)
                if not (mc.hasMatch() and mc.capturedStart() < start):
                    # Definitely a string start. See if it ends on this line
                    me = self._string.match(text, offset=ms.capturedEnd())
                    if me.hasMatch():
                        # Terminate the single-line string
                        end = me.capturedEnd()
                        self.setFormat(start, end - start, STYLE_ANNOT_VAL)
                        strings.append((start, end))
                    else:
                        # Multi-line string
                        self.setCurrentBlockState(1)
                        self.setFormat(start, len(text), STYLE_ANNOT_VAL)

        # Comment
        i = self._comment.globalMatch(text, offset=offset)
        while i.hasNext():
            m = i.next()
            x = m.capturedStart()
            if self._highlight_ok(strings, x, 1):
                self.setFormat(x, len(text) - x, STYLE_COMMENT)
                text = text[:x]
                break

        # Rule-based formatting
        for (pattern, style) in self._rules:
            i = pattern.globalMatch(text, offset=offset)
            while i.hasNext():
                m = i.next()
                x, w = m.capturedStart(), m.capturedLength()
                if self._highlight_ok(strings, x, w):
                    self.setFormat(x, w, style)

        # Model and component headers (must be at start of string)
        if offset == 0:
            m = self._rule_head.match(text)
            if m.hasMatch():
                x, w = m.capturedStart(1), m.capturedLength(1)
                self.setFormat(x, w, STYLE_HEADER)

        # Variable units (must be at start of string)
        if offset == 0:
            m = self._rule_var_unit.match(text)
            if m.hasMatch():
                self.setFormat(
                    m.capturedStart(2), m.capturedLength(2), STYLE_ANNOT_KEY)
                self.setFormat(
                    m.capturedStart(4), m.capturedLength(4), STYLE_ANNOT_VAL)

        # Binds and labels
        for pattern in self._rules_labels:
            i = pattern.globalMatch(text, offset=offset)
            while i.hasNext():
                m = i.next()
                x, w = m.capturedStart(), m.capturedLength()
                if self._highlight_ok(strings, x, w):
                    self.setFormat(m.capturedStart(2), m.capturedLength(2),
                                   STYLE_ANNOT_KEY)
                    self.setFormat(m.capturedStart(3), m.capturedLength(3),
                                   STYLE_ANNOT_VAL)

        # Meta properties (must be at start of string)
        if offset == 0:
            m = self._rule_meta.match(text)
            if m.hasMatch():
                self.setFormat(
                    m.capturedStart(1), m.capturedLength(1), STYLE_ANNOT_KEY)
                # Don't reformat strings (or bits after string end!)
                if m.captured(3)[:3] != '"""':
                    self.setFormat(m.capturedStart(3), m.capturedLength(3),
                                   STYLE_ANNOT_VAL)


class ProtocolHighlighter(QtGui.QSyntaxHighlighter):
    """
    Syntax highlighter for ``mmt`` protocol definitions.
    """
    def __init__(self, document):
        super().__init__(document)

        # Headers and units
        R = QtCore.QRegularExpression
        self._rule_head = R(r'^\s*(\[\[[a-zA-Z0-9_]+\]\])')

        # Highlighting rules
        self._rules = []

        # Numbers
        self._rules.append(
            (R(r'\b[+-]?[0-9]*\.?[0-9]+([eE][+-]?[0-9]+)?\b'), STYLE_LITERAL))

        # Keyword "next"
        self._rules.append((R(r'\bnext\b'), STYLE_KEYWORD_1))

        # Comments
        self._rules.append((R(r'#[^\n]*'), STYLE_COMMENT))

    def highlightBlock(self, text):
        """ Qt: Called whenever a block should be highlighted. """

        # Rule based formatting
        for (pattern, style) in self._rules:
            i = pattern.globalMatch(text)
            while i.hasNext():
                m = i.next()
                self.setFormat(m.capturedStart(), m.capturedLength(), style)

        # Protocol header (must be at strart of string)
        m = self._rule_head.match(text)
        if m.hasMatch():
            self.setFormat(
                m.capturedStart(1), m.capturedLength(1), STYLE_HEADER)


class ScriptHighlighter(QtGui.QSyntaxHighlighter):
    """
    Syntax highlighter for ``mmt`` script files.
    """
    def __init__(self, document):
        super().__init__(document)

        # Script header
        R = QtCore.QRegularExpression
        self._rule_head = R(r'^\s*(\[\[[a-zA-Z0-9_]+\]\])')

        # Highlighting rules
        self._rules = []

        # Keywords
        import keyword
        for kw in keyword.kwlist:
            self._rules.append((R(r'\b' + kw + r'\b'), STYLE_KEYWORD_1))

        # Built-in essential functions
        for func in _PYFUNC:
            self._rules.append((R(r'\b' + str(func) + r'\b'), STYLE_KEYWORD_2))

        # Literals: numbers, True, False, None
        # Override some keywords
        self._rules.append((R(r'\b[+-]?[0-9]*\.?[0-9]+([eE][+-]?[0-9]+)?\b'),
                            STYLE_LITERAL))
        self._rules.append((R(r'\bTrue\b'), STYLE_LITERAL))
        self._rules.append((R(r'\bFalse\b'), STYLE_LITERAL))
        self._rules.append((R(r'\bNone\b'), STYLE_LITERAL))

        # Strings
        self._s1 = R(r'"')
        self._s2 = R(r"'")
        self._ms1 = R(r'"""')
        self._ms2 = R(r"'''")
        self._s_start = R(r'"""|\'\'\'|"|\'')
        self._s_end = {
            '"': self._s1, "'": self._s2, '"""': self._ms1, "'''": self._ms2}

        # Comments
        self._comment = R(r'#')

    def _highlight_ok(self, strings, start, length):
        """ Checks if the string ``start`` to ``length`` needs formatted. """
        for lo, hi in strings:
            if lo <= start < hi or lo <= start + length < hi:
                return False
        return True

    def highlightBlock(self, text):
        """ Qt: Called whenever a block should be highlighted. """

        # To avoid formatting within strings each is stored as a (start, end).
        strings = []
        # If the start has been handled, set the offset.
        offset = 0
        # If the end has been handled, chop it off the string

        # Multi-line strings are done first, because they overrule a lot of
        # things and we can skip formatting if we're inside one.
        # Block states: 0=No string, 1=A " " " string, 2=A ' ' ' string
        self.setCurrentBlockState(0)

        # Continuing a multi-line string?
        previous = self.previousBlockState()
        if previous == 1 or previous == 2:
            # Search for string stop
            r = self._ms1 if previous == 1 else self._ms2
            ms = r.match(text)
            if ms.hasMatch():
                # Terminate the multi-line string, and increase global offset
                offset = ms.capturedEnd(0)
                self.setFormat(0, offset, STYLE_LITERAL)
            else:
                # Whole line in the string
                self.setCurrentBlockState(previous)
                self.setFormat(0, len(text), STYLE_LITERAL)
                return

        # Search for string starts (single or multi-line)
        stroff = offset     # Offset for string start/end searching
        m1 = self._s_start.match(text, offset=stroff)
        while m1.hasMatch():
            stroff = m1.capturedEnd()
            start = m1.capturedStart()
            # Are we in a comment?
            mc = self._comment.match(text)
            if (mc.hasMatch() and mc.capturedStart() < start):
                # No point searching for further string starts
                stroff = len(text)
            else:
                # Find string end
                m2 = self._s_end[m1.captured()].match(text, offset=stroff)
                if m2.hasMatch():
                    stroff = m2.capturedEnd()
                    # Ignore if escaped
                    if text[m2.capturedStart() - 1] != '\\':
                        # Terminate the single line string and move on
                        self.setFormat(start, stroff - start, STYLE_LITERAL)
                        strings.append((start, stroff))
                elif m1.capturedLength() > 1:
                    # Multi-line string start. Block finished!
                    self.setCurrentBlockState(
                        1 if m1.captured() == '"""' else 2)
                    self.setFormat(start, len(text) - start, STYLE_LITERAL)
                    return
                # No Match? Then not a string so ignore and continue
            m1 = self._s_start.match(text, offset=stroff)

        # Comment
        i = self._comment.globalMatch(text, offset=offset)
        while i.hasNext():
            m = i.next()
            x = m.capturedStart()
            if self._highlight_ok(strings, x, 1):
                self.setFormat(x, len(text) - x, STYLE_COMMENT)
                text = text[:x]
                break

        # Script header (must be at start of string)
        if offset == 0:
            m = self._rule_head.match(text)
            if m.hasMatch():
                self.setFormat(
                    m.capturedStart(1), m.capturedLength(1), STYLE_HEADER)

        # Rule based formatting
        for (pattern, style) in self._rules:
            i = pattern.globalMatch(text, offset=offset)
            while i.hasNext():
                m = i.next()
                x, w = m.capturedStart(), m.capturedLength()
                if self._highlight_ok(strings, x, w):
                    self.setFormat(x, w, style)


# List of essential built-in python functions
_PYFUNC = [
    'abs()',
    'aiter()',
    'all()',
    'any()',
    'anext()',
    'ascii()',
    'bin()',
    'bool()',
    'breakpoint()',
    'bytearray()',
    'bytes()',
    'callable()',
    'chr()',
    'classmethod()',
    'compile()',
    'complex()',
    'delattr()',
    'dict()',
    'dir()',
    'divmod()',
    'enumerate()',
    'eval()',
    'exec()',
    'filter()',
    'float()',
    'format()',
    'frozenset()',
    'getattr()',
    'globals()',
    'hasattr()',
    'hash()',
    'help()',
    'hex()',
    'id()',
    'input()',
    'int()',
    'isinstance()',
    'issubclass()',
    'iter()',
    'len()',
    'list()',
    'locals()',
    'map()',
    'max()',
    'memoryview()',
    'min()',
    'next()',
    'object()',
    'oct()',
    'open()',
    'ord()',
    'pow()',
    'print()',
    'property()',
    'range()',
    'repr()',
    'reversed()',
    'round()',
    'set()',
    'setattr()',
    'slice()',
    'sorted()',
    'staticmethod()',
    'str()',
    'sum()',
    'super()',
    'tuple()',
    'type()',
    'vars()',
    'zip()',
    '__import__()',
]



#
# Provides Ansi-C support
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from ._exporter import AnsiCExporter, AnsiCCableExporter, AnsiCEulerExporter
from ._ewriter import AnsiCExpressionWriter


# Importers

# Exporters
_exporters = {
    'ansic': AnsiCExporter,
    'ansic-cable': AnsiCCableExporter,
    'ansic-euler': AnsiCEulerExporter,
}


def exporters():
    """
    Returns a dict of all exporters available in this module.
    """
    return dict(_exporters)


# Expression writers
_ewriters = {
    'ansic': AnsiCExpressionWriter,
}


def ewriters():
    """
    Returns a dict of all expression writers available in this module.
    """
    return dict(_ewriters)


# Language keywords
keywords = [
    # Basic language keywords
    'auto',
    'break',
    'case',
    'char',
    'const',
    'continue',
    'default',
    'do',
    'double',
    'else',
    'enum',
    'extern',
    'float',
    'for',
    'goto',
    'if',
    'inline',
    'int',
    'long',
    'register',
    'return',
    'short',
    'signed',
    'sizeof',
    'static',
    'struct',
    'switch',
    'typedef',
    'union',
    'unsigned',
    'void',
    'volatile',
    'while',
    # From stddef.h
    'NULL',
    'offsetof',
    'ptrdiff_t',
    'wchar_t',
    'size_t',
    # From limits.h
    # TODO?
    # From math.h
    'M_E',
    'M_LOG2E',
    'M_LOG10E',
    'M_LN2',
    'M_LN10',
    'M_PI',
    'M_PI_2',
    'M_PI_4',
    'M_1_PI',
    'M_2_PI',
    'M_2_SQRTPI',
    'M_SQRT2',
    'M_SQRT1_2',
    'MAXFLOAT',
    'HUGE_VAL',
    'acos',
    'asin',
    'atan',
    'atan2',
    'ceil',
    'cos',
    'cosh',
    'exp',
    'fabs',
    'floor',
    'fmod',
    'frexp',
    'ldexp',
    'log',
    'log10',
    'modf',
    'pow',
    'sin',
    'sinh',
    'sqrt',
    'tan',
    'tanh',
    'erf',
    'erfc',
    'gamma',
    'hypot',
    'j0',
    'j1',
    'jn',
    'lgamma',
    'y0',
    'y1',
    'yn',
    'isnan',
    'acosh',
    'asinh',
    'atanh',
    'cbrt',
    'expm1',
    'ilogb',
    'log1p',
    'logb',
    'nextafter',
    'remainder',
    'rint',
    'scalb',
    # From sys/wait.h
    # TODO?
    # From stdlib.h
    # http://pubs.opengroup.org/onlinepubs/7908799/xsh/stdlib.h.html
    # Says: "Inclusion of the <stdlib.h> header may also make visible all
    #        symbols from <stddef.h>, <limits.h>, <math.h> and <sys/wait.h>."
    'EXIT_FAILURE',
    'EXIT_SUCCESS',
    'RAND_MAX',
    'MB_CUR_MAX',
    'div_t',
    'ldiv_t',
    'a64l',
    'abort',
    'abs',
    'atexit',
    'atof',
    'atoi',
    'atol',
    'bsearch',
    'calloc',
    'div',
    'drand48',
    'ecvt',
    'erand48',
    'exit',
    'fcvt',
    'free',
    'gcvt',
    'getenv',
    'getsubopt',
    'grantpt',
    'initstate',
    'jrand48',
    'l64a',
    'labs',
    'lcong48',
    'ldiv',
    'lrand48',
    'malloc',
    'mblen',
    'mbstowcs',
    'mbtowc',
    'mktemp',
    'mkstemp',
    'mrand48',
    'nrand48',
    'ptsname',
    'putenv',
    'qsort',
    'rand',
    'rand_r',
    'random',
    'realloc',
    'realpath',
    'seed48',
    'setkey',
    'setstate',
    'srand',
    'srand48',
    'srandom',
    'strtod',
    'strtol',
    'strtoul',
    'system',
    'ttyslot',
    'unlockpt',
    'valloc',
    'wcstombs',
    'wctomb',
]



#
# This hidden module contains the core functions dealing with simulations and
# the data they generate.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import platform
import sys
import tempfile
import threading
import timeit
import traceback


# Windows fix: On win7 with MinGW, when running distutils from Qt the
# (deprecated) os.popen command fails. The docs suggest to replace calls to
# popen with subprocess.Popen. The following wrapper implements this
# dynamically.
if platform.system() == 'Windows':  # pragma: no linux cover
    import subprocess  # noqa

    def _ospop(command, mode='r', bufsize=0):
        if mode == 'r':
            return subprocess.Popen(
                command, shell=True, bufsize=bufsize, stdout=subprocess.PIPE
            ).stdout
        else:
            return subprocess.Popen(
                command, shell=True, bufsize=bufsize, stdin=subprocess.PIPE
            ).stdin
    os.popen = _ospop

from setuptools import setup, Extension  # noqa

import myokit       # noqa
import myokit.pype  # noqa


def load_module(name, path):
    """ Dynamically find and load a module (Python 3.5+). """
    import importlib
    import importlib.machinery

    spec = importlib.machinery.PathFinder.find_spec(name, [path])
    module = importlib.util.module_from_spec(spec)
    return module


class CModule:
    """
    Abstract base class for classes that dynamically create and compile a
    back-end C-module.
    """
    def __init__(self):
        self._debug_file_count = 0

    def _compile(self, name, template, variables, libs, libd=None, incd=None,
                 carg=None, larg=None, store_build=False,
                 continue_in_debug_mode=False):
        """
        Compiles a source ``template`` with the given ``variables`` into a
        module called ``name``, then imports it and returns a reference to the
        imported module.

        Arguments:

        ``name``
            The name of the generated module (used when importing)
        ``template``
            A template to evaluate
        ``variables``
            Variables to pass in to the template
        ``libs``
            A list of C libraries to link to, e.g.
            ``libs=['sundials_cvodes']``.
        ``libd``
            A list of directories to search for shared library objects, or
            ``None``.
        ``incd``
            A list of directories to search for header files, or ``None``.
        ``carg``
            A list of extra compiler arguments (e.g. ``carg=['-Wall']), or
            ``None``.
        ``larg``
            A list of extra linker arguments (e.g.
            ``larg=['-framework', 'OpenCL']``), or ``None``.
        ``store_build``
            If set to ``False`` (the default), the method will delete the
            temporary directory that the module was built in.
        ``continue_in_debug_mode``
            If ``myokit.DEBUG_SG`` or ``myokit.DEBUG_WG`` are set, the
            generated code will be printed to screen and/or written to disk,
            and ``sys.exit(1)`` will be called. Set ``continue_in_debug_mode``
            to ``True`` to skip the exitting and keep going instead.

        Returns a tuple ``(module, build_path)``, where ``module`` is the
        compiled and imported module, and ``build_path`` is either ``None``
        (the default) or the path to a temporary directory that the build files
        are stored in.

        """
        # Show and/or write code in debug mode
        if myokit.DEBUG_SG or myokit.DEBUG_WG:  # pragma: no cover
            if myokit.DEBUG_SG:
                self._debug_show(template, variables)
            else:
                self._debug_write(template, variables)
            if not continue_in_debug_mode:
                sys.exit(1)

        # Write to temp dir and compile
        src_file = self._source_file()
        working_dir = os.getcwd()
        d_cache = tempfile.mkdtemp('myokit')
        d_build = None
        try:
            # Create build directory
            if store_build:
                # Create separate build directory
                d_build = tempfile.mkdtemp('myokit_build')
            else:
                d_build = os.path.join(d_cache, 'build')
                os.makedirs(d_build)

            # Export c file
            src_file = os.path.join(d_cache, src_file)
            self._export_inner(template, variables, src_file)

            # Ensure headers can be read from myokit/_sim
            if incd is None:
                incd = []
            incd.append(myokit.DIR_CFUNC)

            # Inputs must all be strings
            name = str(name)
            src_file = str(src_file)
            incd = [str(x) for x in incd]
            libd = None if libd is None else [str(x) for x in libd]
            libs = None if libs is None else [str(x) for x in libs]
            carg = None if carg is None else [str(x) for x in carg]
            larg = None if larg is None else [str(x) for x in larg]

            # Show warnings
            if myokit.DEBUG_SC:
                if carg is None:
                    carg = []
                carg.append('-Wall')
                if platform.system() == 'Linux':
                    carg.extend([
                        '-Wextra',
                        '-Wstrict-prototypes',
                        '-Wold-style-definition',
                        #'-Wmissing-prototypes',
                        #'-Wmissing-declarations',
                        '-Wdeclaration-after-statement',
                        '-Wconversion',
                        '-Wno-unused-parameter',
                    ])

            # Add runtime_library_dirs to prevent LD_LIBRARY_PATH errors on
            # unconventional linux sundials installations, but not on windows
            # as this can lead to a weird error in setuptools
            runtime = libd
            if platform.system() == 'Windows':  # pragma: no linux cover
                if libd is not None:
                    runtime = None

                    # Make windows search the libd directories
                    path = os.environ.get('path', '')
                    if path is None:
                        path = ''
                    to_add = [x for x in libd if x not in path]
                    os.environ['path'] = os.pathsep.join([path] + to_add)

                    # In Python 3.8+, they need to be registered with
                    # add_dll_directory too. This does not seem to be 100%
                    # consistent. AppVeyor tests pass when using
                    # add_dll_directory *without* adding the directories to the
                    # path, while installations via miniconda seem to need the
                    # path method too.
                    try:
                        # Fail if add_dll_directory not present
                        os.add_dll_directory

                        # Add DLL paths
                        for path in libd:
                            if os.path.isdir(path):
                                os.add_dll_directory(path)
                    except AttributeError:
                        pass

            # Create extension
            ext = Extension(
                name,
                sources=[src_file],
                libraries=libs,
                library_dirs=libd,
                runtime_library_dirs=runtime,
                include_dirs=incd,
                extra_compile_args=carg,
                extra_link_args=larg,
            )

            # Compile in build directory, catch output
            capture = not (myokit.DEBUG_SC or myokit.COMPAT_NO_CAPTURE)
            fd = not myokit.COMPAT_NO_FD_CAPTURE
            error, trace = None, None
            with myokit.tools.capture(fd=fd, enabled=capture) as s:
                try:
                    os.chdir(d_build)
                    setup(
                        name=name,
                        description='Temporary module',
                        ext_modules=[ext],
                        script_args=[
                            str('build_ext'),
                            str('--inplace'),
                        ])
                except (Exception, SystemExit) as e:  # pragma: no cover
                    error = e
                    trace = traceback.format_exc()
            if error is not None:  # pragma: no cover
                t = ['Unable to compile.', 'Error message:']
                t.append(str(error))
                t.append(trace)
                t.append('Compiler output:')
                captured = s.text().strip()
                t.extend(['    ' + x for x in captured.splitlines()])
                raise myokit.CompilationError('\n'.join(t))

            # Import module
            module = load_module(name, d_build)
            if store_build:
                return module, d_build
            return module

        except Exception:   # pragma: no cover
            # Delete build dir, if created separately
            if store_build and d_build is not None:
                myokit.tools.rmtree(d_build, silent=True)
            raise

        finally:
            # Revert changes to working directory
            os.chdir(working_dir)

            # Delete cache dir (and build dir, if not stored separetely)
            myokit.tools.rmtree(d_cache, silent=True)

    def _debug_show(self, template, variables):  # pragma: no cover
        """ Processes ``template`` and prints the output to screen. """
        print('\n'.join([
            '{:4d}'.format(1 + i) + ' ' + line
            for i, line in enumerate(
                self._export_inner(template, variables).splitlines())
        ]))

    def _debug_write(self, template, variables):  # pragma: no cover
        """ Processes ``template`` and writes the output to file. """
        self._debug_file_count += 1
        fname = 'debug-' + str(self._debug_file_count) + '.c'
        print('DEBUG: Writing generated code to ' + fname)
        with open(fname, 'w') as f:
            f.write(self._export_inner(template, variables))

    def _export(self, template, variables, target=None,
                continue_in_debug_mode=False):
        """
        Exports a source ``template`` with the given ``variables`` and returns
        the result as a string or writes it to the file given by ``target``.

        If ``myokit.DEBUG_SG`` or ``myokit.DEBUG_WG`` are set, the method will
        print the generated code to screen and/or write it to disk. Following
        this, it will terminate with exit code 1 unless
        ``continue_in_debug_mode`` is changed to ``True``.
        """
        # Show and/or write code in debug mode
        if myokit.DEBUG_SG or myokit.DEBUG_WG:  # pragma: no cover
            if myokit.DEBUG_SG:
                self._debug_show(template, variables)
            else:
                self._debug_write(template, variables)
            if not continue_in_debug_mode:
                sys.exit(1)

        return self._export_inner(
            template, variables, target, continue_in_debug_mode)

    def _export_inner(self, template, variables, target=None,
                      continue_in_debug_mode=False):
        """ Internal version of :meth:`_export`. """

        # Test if given module path is writable
        if target is not None:
            if os.path.exists(target):  # pragma: no cover
                # This shouldn't really occur: Writing is always done in a
                # temporary directory
                if os.path.isdir(target):
                    line = 'Can\'t create output file. A directory exists at '
                    line += format_path(target)
                    raise IOError(line)
            # Open output file
            handle = open(target, 'w')

        # Create source
        p = myokit.pype.TemplateEngine()
        if target is not None:
            p.set_output_stream(handle)

        try:
            result = None
            result = p.process(template, variables)
        except myokit.pype.PypeError:  # pragma: no cover
            # Not included in cover, because this can only happen if the
            # template code is wrong, i.e. during development.
            msg = ['An error ocurred while processing the template']
            msg.append(traceback.format_exc())
            d = p.error_details()
            if d:
                msg.append(d)
            raise myokit.GenerationError('\n'.join(msg))
        finally:
            if target is not None:
                handle.close()

        return result

    def _source_file(self):
        """
        Returns a name for the source file created and compiled for this
        module.
        """
        return 'source.c'


class CppModule(CModule):
    """
    Extends the :class:`CModule` class and adds C++ support.
    """
    def _source_file(self):
        return 'source.cpp'


def pid_hash():
    """
    Returns a positive integer hash that depends on the current time as well as
    the process and thread id, so that it's likely to return a different number
    when called twice.
    """
    pid = 1 + os.getpid()                       # Range 0 to 99999
    tid = threading.current_thread().ident      # Non-zero integer
    x = pid * tid * timeit.default_timer()
    return abs(hash(str(x - int(x))))




#
# Converts MathML to Myokit expressions, using an ElementTree implementation.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit

from myokit.formats.xml import split


def parse_mathml_string(s):
    """
    Parses a MathML string that should contain a single expression.
    """
    import xml.etree.ElementTree as etree
    p = MathMLParser(
        lambda x, y: myokit.Name(x),
        lambda x, y: myokit.Number(x),
    )
    return p.parse(etree.fromstring(s))


def parse_mathml_etree(
        element, name_factory, number_factory, free_variables=set()):
    """
    Parses a MathML expression and returns a :class:`myokit.Expression`.

    Arguments:

    ``element``
        An ``xml.etree.ElementTree.Element`` (or similar) to start parsing
        from. Must be an ``<apply>`` element.
    ``name_factory``
        A callable with arguments ``(name_as_string, element)`` that returns
        :class:`myokit.Name` objects.
    ``number_factory``
        A callable with arguments ``(number_as_float, element)`` that returns
        :class:`myokit.Number` objects. Note that ``element`` can be ``None``
        for numbers that have no corresponding ``<cn>`` element.
    ``free_variables``
        All :class:`Name` objects for free variables in derivative expressions
        will be added to this set.

    """
    p = MathMLParser(name_factory, number_factory, free_variables)
    return p.parse(element)


class MathMLError(myokit.ImportError):
    """
    Raised if an error occurs during MathML import.

    The argument ``element`` can be used to pass in an element that caused the
    error.
    """
    def __init__(self, message, element=None):
        if element is not None:
            try:    # pragma: no cover
                line = str(element.sourceline)
                message = 'Error on line ' + line + '. ' + message
            except AttributeError:
                pass
        super().__init__(message)


class MathMLParser:
    """
    Parses MathML expressions into :class:`myokit.Expression` objects.

    Arguments:

    ``name_factory``
        A callable with arguments ``(name_as_string, element)`` that returns
        :class:`myokit.Name` objects.
    ``number_factory``
        A callable with arguments ``(number_as_float, element)`` that returns
        :class:`myokit.Number` objects. Note that ``element`` can be ``None``
        for numbers that have no corresponding ``<cn>`` element.
    ``free_variables``
        All :class:`Name` objects for free variables in derivative expressions
        will be added to this set.

    This is not a validating parser: if the MathML is invalid the method's
    behavior is undefined.

    The following MathML elements are recognised:

    Literals and references

    ``<ci>``
        Is converted to a :class:`myokit.Name` by passing the contents of the
        ``ci`` tag to the ``name_factory``.
    ``<diff>`` (with ``<bvar>`` and ``<degree>``)
        Becomes a :class:`myokit.Derivative`. Only first-order derivatives are
        supported. To check if the derivatives are all time-derivatives, the
        derivative post-processing function can be used.
    ``<cn>``
        Becomes a :class:`myokit.Number`. To process units which may be present
        in the tag's attributes (esp. in CellML) the number post-processing
        function can be used.
    ``<csymbol>``
        Is converted to a :class:`myokit.Name` by passing the contents of its
        ``definitionURL`` to the ``name_factory``. Note that ``csymbols``
        representing operators or functions are not supported.

    Algebra

    ``<plus>``
        Becomes a :class:`myokit.PrefixPlus`, a :class`myokit.Plus` or a tree
        of :class:`myokit.Plus` elements.
    ``<minus>``
        Becomes a :class:`myokit.PrefixMinus`, a :class`myokit.Minus` or a tree
        of :class:`myokit.Minus` elements.
    ``<times>``
        Becomes a :class:`myokit.Multiply` or a tree of
        :class:`myokit.Multiply` elements.
    ``<divide>``
        Becomes a :class:`myokit.Divide` or a tree of :class:`myokit.Divide`
        elements.
    ``<apply>``
        Used to indicate the tree structure of the equation. These get
        translated but don't have a Myokit counterpart.

    Functions

    ``<power>``
        Becomes a :class:`myokit.Power`.
    ``<root>`` (with ``<degree>``)
        Becomes a :class:`myokit.Sqrt`.
    ``<exp>``
        Becomes a :class:`myokit.Exp`.
    ``<ln>``
        Becomes a :class:`myokit.Log`.
    ``<log>`` (with ``<logbase>``)
        Becomes a :class:`myokit.Log10` or a :class:`myokit.Log`.
    ``<abs>``
        Becomes a :class:`myokit.Abs`.
    ``<floor>``
        Becomes a :class:`myokit.Floor`.
    ``<ceiling>``
        Becomes a :class:`myokit.Ceil`.
    ``<quotient>``
        Becomes a :class:`myokit.Quotient`.
    ``<rem>``
        Becomes a :class:`myokit.Remainder`.

    Trigonometry

    ``<sin>``, ``<cos>`` and ``<tan>``
        Become :class:`myokit.Sin`, :class:`myokit.Cos` and
        :class:`myokit.Tan`.
    ``<arcsin>``, ``<arccos>`` and ``<arctan>``
        Become :class:`myokit.ASin`, :class:`myokit.ACos` and
        :class:`myokit.ATan`.
    ``<csc>``, ``<sec>`` and ``<cot>``
        Become ``1/sin``, ``1/cos`` and ``1/tan``.
    ``<arccsc>``, ``<arcsec>`` and ``<arccot>``
        Become ``asin(1/x)``, ``acos(1/x)`` and ``atan(1/x)``.

    Hyperbolic trigonometry

    ``<sinh>``
        Becomes ``0.5 * (exp(x) - exp(-x))``.
    ``<cosh>``
        Becomes ``0.5 * (exp(x) + exp(-x))``.
    ``<tanh>``
        Becomes ``(exp(2 * x) - 1) / (exp(2 * x) + 1)``.
    ``<arcsinh>``
        Becomes ``log(x + sqrt(x*x + 1))``.
    ``<arccosh>``
        Becomes ``log(x + sqrt(x*x - 1))``.
    ``<arctanh>``
        Becomes ``0.5 * log((1 + x) / (1 - x))``.
    ``<csch>``
        Becomes ``2 / (exp(x) - exp(-x))``.
    ``<sech>``
        Becomes ``2 / (exp(x) + exp(-x))``.
    ``<coth>``
        Becomes ``(exp(2 * x) + 1) / (exp(2 * x) - 1)``.
    ``<arccsch>``
        Becomes ``log(1 / x + sqrt(1 / x^2 + 1))``.
    ``<arcsech>``
        Becomes ``log(1 / x + sqrt(1 / x^2 - 1))``
    ``<arccoth>``
        Becomes ``0.5 * log((x + 1) / (x - 1))``.

    Logic and relations

    ``<piecewise>``, ``<piece>`` and ``<otherwise>``
        Becomes a :class:`myokit.Piecewise`.
    ``<and>``, ``<or>`` and ``<not>``
        Become :class:`myokit.And`, :class:`myokit.Or` and :class:`myokit.Not`.
    ``<xor>``
        Becomes ``(x or y) and not(x and y)``
    ``<eq>`` and ``<neq>``
        Becomes :class:`myokit.Equal` and :class:`NotEqual`.
    ``<lt>`` and ``<gt>``
        Become :class:`myokit.Less` and :class:`myokit.More`.
    ``<leq>`` and ``<geq>``
        Become :class:`myokit.LessEqual` and :class:`myokit.MoreEqual`.

    Constants

    ``<pi>``
        Becomes ``3.14159265358979323846``
    ``<exponentiale>``
        Becomes ``exp(1)``
    ``<true>``
        Becomes ``1``
    ``<false>``
        Becomes ``0``

    """
    def __init__(self, variable_factory, number_factory, free_variables=set()):
        self._vfac = variable_factory
        self._nfac = number_factory
        self._const = lambda x: number_factory(x, None)
        self._free_variables = free_variables

    def _eat(self, element, iterator, nargs=1):
        """
        Takes ``nargs`` elements from the ``iterator``, which should be
        pointing at the first element after ``element``.

        Will complain if there are fewer or more than ``nargs`` elements
        available from the iterator.

        ``element``
            The element just before the iterator (used in error messages).
        ``iterator``
            An iterator pointing at the first element after ``element``.
        ``nargs``
            The required number of arguments that ``iterator`` should still
            contain.

        """
        # Get all operands
        ops = [self._parse_atomic(x) for x in iterator]

        # Check number of operands
        if len(ops) != nargs:
            raise MathMLError(
                'Expecting ' + str(nargs) + ' operand(s), got ' + str(len(ops))
                + ' for ' + split(element.tag)[1] + '.', element)

        return ops

    def _next(self, iterator, tag=None):
        """
        Returns the next element from an ``iterator``.

        If ``tag`` is given, elements will be drawn from the iterator until an
        element with the given tag is found.
        """
        try:
            # Return next element
            if tag is None:
                return next(iterator)

            # Find a specific element
            el = next(iterator)
            while split(el.tag)[1] != tag:
                el = next(iterator)
            return el

        # Ran out of elements
        except StopIteration:
            return None

    def parse(self, element):
        """
        Parses a MathML expression, rooted in the given ``<apply>`` element,
        and returns a :class:`myokit.Expression`.

        Arguments:

        ``element``
            An ``xml.etree.ElementTree.Element`` (or similar) to start parsing
            from. Must be an ``<apply>`` element.

        """
        # Remove <math> element, if found
        ns, el = split(element.tag)
        if el == 'math':
            element = element[0]

        return self._parse_atomic(element)

    def _parse_atomic(self, element):
        """
        Parses a bit of MathML entirely encoded in ``element``, i.e. a number,
        variable, or apply.
        """

        # Get element type, decide what to do
        _, name = split(element.tag)

        # Brackets
        if name == 'apply':
            return self._parse_apply(element)

        # Variable reference
        elif name == 'ci':
            return self._parse_name(element)

        # Number
        elif name == 'cn':
            return self._parse_number(element)

        elif name == 'csymbol':
            return self._parse_symbol(element)

        # Constants
        elif name == 'pi':
            return self._const('3.14159265358979323846')
        elif name == 'exponentiale':
            return myokit.Exp(self._const(1))
        elif name == 'true':
            # This is correct, ``True == 1`` -> False, ``True == 2`` -> False
            return self._const(1)
        elif name == 'false':
            return self._const(0)
        elif name == 'notanumber':
            return self._const(float('nan'))
        elif name == 'infinity':
            return self._const(float('inf'))

        # Piecewise statement
        elif name == 'piecewise':
            return self._parse_piecewise(element)

        # Unexpected element
        else:
            raise MathMLError(
                'Unsupported element: ' + str(element.tag) + '.', element)

    def _parse_apply(self, apply_element):
        """
        Parses an ``<apply>`` element.
        """
        # Apply must have kids
        if len(apply_element) == 0:
            raise MathMLError(
                'Apply must contain at least one child element.',
                apply_element)

        # Get first child
        iterator = iter(apply_element)
        element = self._next(iterator)

        # Decide what to do based on first child
        _, name = split(element.tag)

        # Handle derivative
        if name == 'diff':
            return self._parse_derivative(element, iterator)

        # Algebra (unary/binary/n-ary operators)
        elif name == 'plus':
            return self._parse_nary(
                element, iterator, myokit.Plus, myokit.PrefixPlus)
        elif name == 'minus':
            return self._parse_nary(
                element, iterator, myokit.Minus, myokit.PrefixMinus)
        elif name == 'times':
            return self._parse_nary(element, iterator, myokit.Multiply)
        elif name == 'divide':
            return self._parse_nary(element, iterator, myokit.Divide)

        # Basic functions
        elif name == 'exp':
            return myokit.Exp(*self._eat(element, iterator))
        elif name == 'ln':
            return myokit.Log(*self._eat(element, iterator))
        elif name == 'log':
            return self._parse_log(element, iterator)
        elif name == 'root':
            return self._parse_root(element, iterator)
        elif name == 'power':
            return myokit.Power(*self._eat(element, iterator, 2))
        elif name == 'floor':
            return myokit.Floor(*self._eat(element, iterator))
        elif name == 'ceiling':
            return myokit.Ceil(*self._eat(element, iterator))
        elif name == 'abs':
            return myokit.Abs(*self._eat(element, iterator))
        elif name == 'quotient':
            return myokit.Quotient(*self._eat(element, iterator, 2))
        elif name == 'rem':
            return myokit.Remainder(*self._eat(element, iterator, 2))

        # Logic
        elif name == 'and':
            return self._parse_nary(element, iterator, myokit.And)
        elif name == 'or':
            return self._parse_nary(element, iterator, myokit.Or)
        elif name == 'xor':
            # Becomes ``(x or y) and not(x and y)``
            x, y = self._eat(element, iterator, 2)
            return myokit.And(myokit.Or(x, y), myokit.Not(myokit.And(x, y)))

        elif name == 'not':
            return myokit.Not(*self._eat(element, iterator))
        elif name == 'eq' or name == 'equivalent':
            return myokit.Equal(*self._eat(element, iterator, 2))
        elif name == 'neq':
            return myokit.NotEqual(*self._eat(element, iterator, 2))
        elif name == 'gt':
            return myokit.More(*self._eat(element, iterator, 2))
        elif name == 'lt':
            return myokit.Less(*self._eat(element, iterator, 2))
        elif name == 'geq':
            return myokit.MoreEqual(*self._eat(element, iterator, 2))
        elif name == 'leq':
            return myokit.LessEqual(*self._eat(element, iterator, 2))

        # Trigonometry
        elif name == 'sin':
            return myokit.Sin(*self._eat(element, iterator))
        elif name == 'cos':
            return myokit.Cos(*self._eat(element, iterator))
        elif name == 'tan':
            return myokit.Tan(*self._eat(element, iterator))
        elif name == 'arcsin':
            return myokit.ASin(*self._eat(element, iterator))
        elif name == 'arccos':
            return myokit.ACos(*self._eat(element, iterator))
        elif name == 'arctan':
            return myokit.ATan(*self._eat(element, iterator))

        # Redundant trigonometry (CellML includes this)
        elif name == 'csc':
            # Cosecant: csc(x) = 1 / sin(x)
            return myokit.Divide(
                self._const(1), myokit.Sin(*self._eat(element, iterator)))
        elif name == 'sec':
            # Secant: sec(x) = 1 / cos(x)
            return myokit.Divide(
                self._const(1), myokit.Cos(*self._eat(element, iterator)))
        elif name == 'cot':
            # Contangent: cot(x) = 1 / tan(x)
            return myokit.Divide(
                self._const(1), myokit.Tan(*self._eat(element, iterator)))
        elif name == 'arccsc':
            # ArcCosecant: acsc(x) = asin(1/x)
            return myokit.ASin(
                myokit.Divide(self._const(1), *self._eat(element, iterator)))
        elif name == 'arcsec':
            # ArcSecant: asec(x) = acos(1/x)
            return myokit.ACos(
                myokit.Divide(self._const(1), *self._eat(element, iterator)))
        elif name == 'arccot':
            # ArcCotangent: acot(x) = atan(1/x)
            return myokit.ATan(
                myokit.Divide(self._const(1), *self._eat(element, iterator)))

        # Hyperbolic trig
        elif name == 'sinh':
            # Hyperbolic sine: sinh(x) = 0.5 * (e^x - e^-x)
            x = self._eat(element, iterator)[0]
            return myokit.Multiply(
                self._const(0.5), myokit.Minus(
                    myokit.Exp(x), myokit.Exp(myokit.PrefixMinus(x))))
        elif name == 'cosh':
            # Hyperbolic cosine: cosh(x) = 0.5 * (e^x + e^-x)
            x = self._eat(element, iterator)[0]
            return myokit.Multiply(
                self._const(0.5), myokit.Plus(
                    myokit.Exp(x), myokit.Exp(myokit.PrefixMinus(x))))
        elif name == 'tanh':
            # Hyperbolic tangent: tanh(x) = (e^2x - 1) / (e^2x + 1)
            x = self._eat(element, iterator)[0]
            e2x = myokit.Exp(myokit.Multiply(self._const(2), x))
            return myokit.Divide(
                myokit.Minus(e2x, self._const(1)),
                myokit.Plus(e2x, self._const(1)))
        elif name == 'arcsinh':
            # Inverse hyperbolic sine: asinh(x) = log(x + sqrt(x*x + 1))
            x = self._eat(element, iterator)[0]
            return myokit.Log(myokit.Plus(x, myokit.Sqrt(myokit.Plus(
                myokit.Multiply(x, x), self._const(1)))))
        elif name == 'arccosh':
            # Inverse hyperbolic cosine:
            #   acosh(x) = log(x + sqrt(x*x - 1))
            x = self._eat(element, iterator)[0]
            return myokit.Log(myokit.Plus(x, myokit.Sqrt(myokit.Minus(
                myokit.Multiply(x, x), self._const(1)))))
        elif name == 'arctanh':
            # Inverse hyperbolic tangent:
            #   atanh(x) = 0.5 * log((1 + x) / (1 - x))
            x = self._eat(element, iterator)[0]
            return myokit.Multiply(
                self._const(0.5), myokit.Log(myokit.Divide(
                    myokit.Plus(self._const(1), x),
                    myokit.Minus(self._const(1), x))))

        # Hyperbolic redundant trig
        elif name == 'csch':
            # Hyperbolic cosecant: csch(x) = 2 / (exp(x) - exp(-x))
            x = self._eat(element, iterator)[0]
            return myokit.Divide(
                self._const(2), myokit.Minus(
                    myokit.Exp(x), myokit.Exp(myokit.PrefixMinus(x))))
        elif name == 'sech':
            # Hyperbolic secant: sech(x) = 2 / (exp(x) + exp(-x))
            x = self._eat(element, iterator)[0]
            return myokit.Divide(
                self._const(2), myokit.Plus(
                    myokit.Exp(x), myokit.Exp(myokit.PrefixMinus(x))))
        elif name == 'coth':
            # Hyperbolic cotangent:
            #   coth(x) = (exp(2*x) + 1) / (exp(2*x) - 1)
            x = self._eat(element, iterator)[0]
            e2x = myokit.Exp(myokit.Multiply(self._const(2), x))
            return myokit.Divide(
                myokit.Plus(e2x, self._const(1)),
                myokit.Minus(e2x, self._const(1)))
        elif name == 'arccsch':
            # Inverse hyperbolic cosecant:
            #   arccsch(x) = log(1 / x + sqrt(1 / x^2 + 1))
            x = self._eat(element, iterator)[0]
            return myokit.Log(myokit.Plus(
                myokit.Divide(self._const(1), x),
                myokit.Sqrt(myokit.Plus(
                    myokit.Divide(self._const(1), myokit.Multiply(x, x)),
                    self._const(1)))))
        elif name == 'arcsech':
            # Inverse hyperbolic secant:
            #   arcsech(x) = log(1 / x + sqrt(1 / x^2 - 1))
            x = self._eat(element, iterator)[0]
            return myokit.Log(myokit.Plus(
                myokit.Divide(self._const(1), x),
                myokit.Sqrt(myokit.Minus(
                    myokit.Divide(self._const(1), myokit.Multiply(x, x)),
                    self._const(1)))))
        elif name == 'arccoth':
            # Inverse hyperbolic cotangent:
            #   arccoth(x) = 0.5 * log((x + 1) / (x - 1))
            x = self._eat(element, iterator)[0]
            return myokit.Multiply(
                self._const(0.5), myokit.Log(myokit.Divide(
                    myokit.Plus(x, self._const(1)),
                    myokit.Minus(x, self._const(1)))))

        # Last option: A single atomic inside an apply
        # Do this one last to stop e.g. <apply><times /></apply> returning the
        # error 'Unsupported element' (which is what parse_atomic would call).
        elif len(apply_element) == 1:
            return self._parse_atomic(element)

        # Unexpected element
        else:
            raise MathMLError(
                'Unsupported element in apply: ' + str(element.tag) + '.',
                element)

    def _parse_derivative(self, element, iterator):
        """
        Parses the elements folling a ``<diff>`` element.

        Arguments

        ``element``
            A ``<diff>`` element
        ``iterator``
            An iterator pointing at the next element.

        """

        # Get free variable
        bvar = self._next(iterator, 'bvar')
        if bvar is None:
            raise MathMLError(
                '<diff> element must contain a <bvar>.', element)
        ci = self._next(iter(bvar), 'ci')
        if ci is None:
            raise MathMLError(
                '<bvar> element must contain a <ci>', element)
        self._free_variables.add(self._parse_name(ci))

        # Check degree, if given
        degree = self._next(iter(bvar), 'degree')
        if degree is not None:
            cn = self._next(iter(degree), 'cn')
            if cn is None:
                raise MathMLError(
                    '<degree> element must contain a <cn>.', degree)
            d = self._parse_number(cn)
            if d.eval() != 1:
                raise MathMLError(
                    'Only derivatives of degree one are supported.', cn)

        # Get Name object
        ci = self._next(iterator, 'ci')
        if ci is None:
            raise MathMLError(
                '<diff> element must contain a <ci> after its <bvar>'
                ' element (derivatives of expressions are not supported.',
                element)
        var = self._parse_name(ci)

        return myokit.Derivative(var)

    def _parse_log(self, element, iterator):
        """
        Parses the elements following a ``<log>`` element.

        Arguments:

        ``element``
            The ``<log>`` element.
        ``iterator``
            An iterator pointing at the first element after ``element``.

        """
        # Get next operands
        ops = [x for x in iterator]

        # Check for zero ops
        if len(ops) == 0:
            raise MathMLError(
                'Expecting operand after <log> element.', element)

        # Check if first op is logbase
        if split(ops[0].tag)[1] == 'logbase':

            # Get logbase
            base = ops[0]
            if len(base) != 1:
                raise MathMLError(
                    'Expecting a single operand inside <logbase> element.',
                    base)
            base = self._parse_atomic(base[0])

            # Get main operand
            if len(ops) != 2:
                raise MathMLError(
                    'Expecting a single operand after the <logbase> element'
                    ' inside a <log>.', element)
            op = self._parse_atomic(ops[1])

        # No logbase given
        else:
            base = None

            if len(ops) != 1:
                raise MathMLError(
                    'Expecting a single operand (or a <logbase> followed by a'
                    ' single operand) inside a <log> element.', element)
            op = self._parse_atomic(ops[0])

        if base is None or base.eval() == 10:
            return myokit.Log10(op)
        else:
            return myokit.Log(op, base)

    def _parse_nary(self, element, iterator, binary, unary=None):
        """
        Parses operands for unary, binary, or n-ary operators (for example
        plus, minus, times and division).

        If n-ary expressions (n > 2) are encountered, these are converted to
        Myokit expression trees.

        Arguments:

        ``element``
            The element that determined the operator type.
        ``iterator``
            An iterator pointing at the first element after ``element``.
        ``binary``
            A ``myokit.Expression`` subclass for a binary expression.
        ``unary``
            An optional ``myokit.Expression`` subclass for unary operators.

        """
        # Get all operands
        ops = [self._parse_atomic(x) for x in iterator]

        # Check the number of operands
        n = len(ops)
        if n < 1:
            raise MathMLError(
                'Operator needs at least one operand.', element)
        if n < 2:
            if unary:
                return unary(ops[0])
            else:
                raise MathMLError(
                    'Operator needs at least two operands', element)

        # Create nested binary expressions and return
        ex = binary(ops[0], ops[1])
        for i in range(2, n):
            ex = binary(ex, ops[i])
        return ex

    def _parse_number(self, element):
        """
        Parses a ``<cn>`` element and returns a number object created by the
        number factory.
        """
        # https://www.w3.org/TR/MathML2/chapter4.html#contm.typeattrib
        kind = element.attrib.get('type', 'real')

        # Get value
        if kind == 'real':
            # Float, specified as 123.123 (no exponent!)
            # May be in a different base than 10
            base = element.attrib.get('base', '10').strip()
            try:
                base = float(base)
            except (TypeError, ValueError):
                raise MathMLError(
                    'Invalid base specified on <ci> element.', element)
            if base != 10:
                raise MathMLError(
                    'Numbers in bases other than 10 are not supported.',
                    element)

            # Get value
            # Note: We are being tolerant here and allowing e-notation (which
            # is not consistent with the spec!)
            if element.text is None:
                raise MathMLError('Empty <cn> element', element)
            try:
                value = float(element.text.strip())
            except ValueError:
                raise MathMLError(
                    'Unable to convert contents of <cn> to a real number: "'
                    + str(element.text) + '"', element)

        elif kind == 'integer':
            # Integer in any given base
            base = element.attrib.get('base', '10').strip()
            try:
                base = int(base)
            except ValueError:
                raise MathMLError(
                    'Unable to parse base of <cn> element: "' + base + '"',
                    element)

            # Get value
            if element.text is None:
                raise MathMLError('Empty <cn> element', element)
            try:
                value = int(element.text.strip(), base)
            except ValueError:
                raise MathMLError(
                    'Unable to convert contents of <cn> to an integer: "'
                    + str(element.text) + '"', element)

        elif kind == 'double':
            # Floating point (positive, negative, exponents, etc)

            if element.text is None:
                raise MathMLError('Empty <cn> element', element)
            try:
                value = float(element.text.strip())
            except ValueError:
                raise MathMLError(
                    'Unable to convert contents of <cn> to a real number: "'
                    + str(element.text) + '"', element)

        elif kind == 'e-notation':
            # 1<sep />3 = 1e3

            # Check contents
            parts = [x for x in element]
            if len(parts) != 1 or split(parts[0].tag)[1] != 'sep':
                raise MathMLError(
                    'Number in e-notation should have the format'
                    ' number<sep />number.', element)

            # Get parts of number
            sig = element.text
            exp = parts[0].tail
            if sig is None or not sig.strip():
                raise MathMLError(
                    'Unable to parse number in e-notation: missing part before'
                    ' the separator.', element)
            if exp is None or not exp.strip():
                raise MathMLError(
                    'Unable to parse number in e-notation: missing part after'
                    ' the separator.', element)

            # Get value
            try:
                value = float(sig.strip() + 'e' + exp.strip())
            except ValueError:
                raise MathMLError(
                    'Unable to parse number in e-notation "' + sig + 'e' + exp
                    + '".', element)

        elif kind == 'rational':
            # 1<sep />3 = 1 / 3
            # Check contents
            parts = [x for x in element]
            if len(parts) != 1 or split(parts[0].tag)[1] != 'sep':
                raise MathMLError(
                    'Rational number should have the format'
                    ' number<sep />number.', element)

            # Get parts of number
            numer = element.text
            denom = parts[0].tail
            if numer is None or not numer.strip():
                raise MathMLError(
                    'Unable to parse rational number: missing part before the'
                    ' separator.', element)
            if denom is None or not denom.strip():
                raise MathMLError(
                    'Unable to parse rational number: missing part after the'
                    ' separator.', element)

            # Get value
            try:
                value = float(numer.strip()) / float(denom.strip())
            except ValueError:
                raise MathMLError(
                    'Unable to parse rational number "' + numer + ' / ' + denom
                    + '".', element)

        else:
            raise MathMLError('Unsupported <cn> type: ' + kind, element)

        # Create number and return
        return self._nfac(value, element)

    def _parse_piecewise(self, element):
        """
        Parses a ``<piecewise>`` element.
        """

        # Piecewise contains at least one piece, optionally contains an
        # "otherwise". Syntax doesn't ensure this statement makes sense.
        ops = []
        other = None

        # Scan pieces
        for child in element:
            _, el = split(child.tag)

            if el == 'piece':
                if len(child) != 2:
                    raise MathMLError(
                        '<piece> element must have exactly 2 children.', child)
                ops.append(self._parse_atomic(child[1]))    # Condition
                ops.append(self._parse_atomic(child[0]))    # Value

            elif el == 'otherwise':
                if other is not None:
                    raise MathMLError(
                        'Found more than one <otherwise> inside a <piecewise>'
                        ' element.', child)
                if len(child) != 1:
                    raise MathMLError(
                        '<otherwise> element must have exactly 1 child.',
                        child)
                other = self._parse_atomic(child[0])

            else:
                raise MathMLError(
                    'Unexpected content in <piecewise>. Expecting <piece> or'
                    ' <otherwise>, found <' + el + '>.', child)

        # Add otherwise
        if other is None:
            ops.append(self._const(0))
        else:
            ops.append(other)

        return myokit.Piecewise(*ops)

    def _parse_root(self, element, iterator):
        """
        Parses the elements following a ``<root>`` element.

        Arguments:

        ``element``
            The ``<root>`` element.
        ``iterator``
            An iterator pointing at the first element after ``element``.

        """
        # Get next operands
        ops = [x for x in iterator]

        # Check for zero ops
        if len(ops) == 0:
            raise MathMLError(
                'Expecting operand after <root> element.', element)

        # Check if first op is degree
        if split(ops[0].tag)[1] == 'degree':

            # Get degree
            degree = ops[0]
            if len(degree) != 1:
                raise MathMLError(
                    'Expecting a single operand inside <degree> element.',
                    degree)
            degree = self._parse_atomic(degree[0])

            # Get main operand
            if len(ops) != 2:
                raise MathMLError(
                    'Expecting a single operand after the <degree> element'
                    ' inside a <root>.', element)
            op = self._parse_atomic(ops[1])

            # Return expression
            if degree.eval() == 2:
                return myokit.Sqrt(op)
            return myokit.Power(op, myokit.Divide(self._const(1), degree))

        # No degree given
        if len(ops) != 1:
            raise MathMLError(
                'Expecting a single operand (or a <degree> followed by a'
                ' single operand) inside a <root> element.', element)
        op = self._parse_atomic(ops[0])

        return myokit.Sqrt(op)

    def _parse_name(self, element):
        """
        Parses a ``<ci>`` element and returns a :class:`myokit.Name` created by
        the name factory.
        """
        if element.text is None:
            raise MathMLError(
                '<ci> element must contain a variable name.', element)

        symbol = element.text.strip()
        try:
            return self._vfac(symbol, element)
        except Exception as e:
            raise MathMLError('Unable to create Name: ' + str(e), element)

    def _parse_symbol(self, element):
        """
        Parses only ``<csymbol>`` elements that represent special variables
        and returns a :class:`myokit.Name` created by the name factory.
        """
        symbol = element.get('definitionURL')
        if symbol is None:
            raise MathMLError(
                '<csymbol> element must contain a definitionURL attribute.',
                element)

        symbol = symbol.strip()
        try:
            return self._vfac(symbol, element)
        except Exception as e:
            raise MathMLError(
                'Unable to create Name from csymbol: ' + str(e), element)




#
# C++ expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from myokit.formats.ansic import AnsiCExpressionWriter


class CppExpressionWriter(AnsiCExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    myokit :class:`expressions <myokit.Expression>` to their C++ equivalent.
    """
    pass



#
# Myokit utility functions. This module gathers functions that are used in
# Myokit, but are not particularly Myokit-dependent, i.e. they could easily be
# stand-alone. Imported by default.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import io
import os
import re
import shutil
import stat
import sys
import tempfile
import threading
import timeit


# Natural sort regex
_natural_sort_regex = re.compile('([0-9]+)')


class Benchmarker:
    """
    Allows benchmarking using the with statement.

    Example::

        m,p,x = myokit.load('example')
        s = myokit.Simulation(m, p)
        b = myokit.tools.Benchmarker()
        s.run()
        print(b.time())
        b.reset()
        s.run()
        print(b.time())

    """
    def __init__(self):
        self._start = timeit.default_timer()
        self._last_print = self._start

    def format(self, time=None):
        """
        Formats a (non-integer) number of seconds, returns a string like
        "5 weeks, 3 days, 1 hour, 4 minutes, 9 seconds", or "0.0019 seconds".

        If no ``time`` is passed in, the value from :meth:`time()` is used.
        """
        if time is None:
            time = self.time()
        if time < 60:
            return '1 second' if time == 1 else str(time) + ' seconds'
        output = []
        time = int(round(time))
        units = [
            (604800, 'week'),
            (86400, 'day'),
            (3600, 'hour'),
            (60, 'minute'),
        ]
        for k, name in units:
            f = time // k
            if f > 0 or output:
                output.append(str(f) + ' ' + (name if f == 1 else name + 's'))
            time -= f * k
        output.append('1 second' if time == 1 else str(time) + ' seconds')
        return ', '.join(output)

    def print(self, message):
        """
        Prints a message to stdout, preceded by the benchmarker time in us.
        """
        now = timeit.default_timer()
        tot = int(1e6 * (now - self._start))
        new = int(1e6 * (now - self._last_print))
        self._last_print = now
        print('[{:10d} us ({:5d} us)] '.format(tot, new) + str(message))

    def reset(self):
        """ Resets this timer's start time. """
        self._start = timeit.default_timer()

    def time(self):
        """
        Returns the time since benchmarking started (as a float, in seconds).
        """
        return timeit.default_timer() - self._start


class capture:
    """
    Context manager that temporarily redirects the current standard output and
    error streams, and captures anything that's written to them.

    Example::

        with myokit.tools.capture() as a:
            print('This will be captured')

    Within a single thread, captures can be nested, for example::

        with myokit.tools.capture() as a:
            print('This will be captured by a')
            with myokit.tools.capture() as b:
                print('This will be captured by b, not a')
            print('This will be captured by a again')

    Capturing is thread-safe: a lock is used to ensure only a single thread is
    capturing at any time. For example, if we have a function::

        def f(i):
            with myokit.tools.capture() as a:
                print(i)
                ...

            return a.text()

    and this is called from several threads, the ``capture`` acts as a lock (a
    ``threading.RLock``) so that one thread will need to finish executing the
    code within the ``with`` statement before a second thread can start
    capturing.

    In multiprocessing, no locks are used, and no memory or streams are shared
    so that this should also be safe.

    By default, this method works by simply redirecting ``sys.stdout`` and
    ``sys.stderr``. This captures any output written by the Python interpreter,
    but does not catch output from C/C++ extensions or subrocesses. To also
    catch that output start the capture with the optional argument
    ``fd=True``, which enables a file descriptor duplication method of
    redirection.

    To easily switch capturing on/off, a switch ``enabled=False`` can be passed
    in to create a context manager that doesn't do anything.
    """
    # Note: It seems we need to capture both streams to make the file
    # descriptor method work, and we want both anyway throughout Myokit, so
    # there are no options to choose stdout or stderr here.

    # Lock to stop other threads from capturing while this thread is capturing.
    _rlock = threading.RLock()

    def __init__(self, fd=False, enabled=True):

        # Are we already capturing? This is needed in case someone enters the
        # same context twice.
        self._active_count = 0

        # Captured text
        self._txt_out = None
        self._txt_err = None

        # Capturing mode
        self._fd = bool(fd)

        # Shared by both modes
        self._org_out = None     # Original stdout object
        self._org_err = None     # Original stderr object

        # Python stream redirects only
        self._tmp_out = None    # Temporary stdout StringIO object
        self._tmp_err = None    # Temporary stderr StringIO object

        # File descriptor redirects only
        self._out_fd = None      # File descriptor used for output
        self._err_fd = None      # File descriptor used for errors
        self._org_out_fd = None  # Back-up of original stdout file descriptor
        self._org_err_fd = None  # Back-up of original stderr file descriptor
        self._file_out = None    # Temporary file to write output to
        self._file_err = None    # Temporary file to write errors to

        # Capturing enabled
        self._enabled = bool(enabled)

    def __enter__(self):
        """Called when the context is entered."""
        if not self._enabled:
            return self

        # Avoid entering the same context object twice
        self._active_count += 1
        if self._active_count == 1:

            # Wait until other threads have stopped capturing
            capture._rlock.acquire()

            # Set up redirection
            self._start()

        # Return context
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Called when exiting the context."""
        if not self._enabled:
            return

        self._active_count -= 1
        if self._active_count == 0:
            self._stop()
            capture._rlock.release()

    def _start(self):
        """Starts capturing output to stdout and stderr."""

        # Clear any cached text
        self._txt_out = None
        self._txt_err = None

        # Save the current output / error streams
        self._org_out = sys.stdout
        self._org_err = sys.stderr

        # Redirect
        if not self._fd:
            # Create temporary output and error streams
            self._tmp_out = io.StringIO()
            self._tmp_err = io.StringIO()

            # Redirect, attempting to flush first
            try:
                sys.stdout.flush()
            except AttributeError:  # pragma: no cover
                pass
            finally:
                sys.stdout = self._tmp_out

            try:
                sys.stderr.flush()
            except AttributeError:  # pragma: no cover
                pass
            finally:
                sys.stderr = self._tmp_err

        else:
            # File-descriptor redirection

            # Get file descriptors used for output and errors.
            #
            # These represent open files, and are the low-level equivalent of
            # sys.stdout and sys.stderr: all subprocesses etc. use these
            # descriptors to write output to.
            #
            # Note that we get these file descriptors from __stdout__ and
            # __stderr__, to get the original streams that subprocesses etc.
            # will use, even if someone has redirected the Python-level
            # streams.
            #
            # On https://docs.python.org/3/library/sys.html#module-sys, it says
            # that stdout/err as well as __stdout__ can be None (e.g. in spyder
            # on windows).  In other cases (pythonw.exe) they can be set but
            # return a negative file descriptor (indicating it's invalid).
            # So here we check if __stdout__ is None and if so set a negative
            # fileno so that we can catch both cases at once in the rest of the
            # code.
            #
            if sys.__stdout__ is not None:
                self._out_fd = sys.__stdout__.fileno()
            else:   # pragma: no cover
                self._out_fd = -1
            if sys.__stderr__ is not None:
                self._err_fd = sys.__stderr__.fileno()
            else:   # pragma: no cover
                self._err_fd = -1

            # We start by making a copy of these file descriptors using dup(),
            # meaning that two file numbers will point to the same file. In a
            # later step we use dup2() to write a different value into the
            # descriptor, so that all output gets redirected. When exiting, we
            # will use dup2() again to copy the original value back in.
            #
            # We check if the fd >= 0 first, because the previous code could
            # get a negative (invalid) value either when __stdxxx__ is None or
            # when a negative value is returned.
            #
            # On windows, the order is important: First dup both stdout and
            # stderr, then dup2 the new descriptors in. This prevents a weird
            # infinite recursion on windows ipython / python shell.
            #
            if self._out_fd >= 0:
                self._org_out_fd = os.dup(self._out_fd)
            if self._err_fd >= 0:
                self._org_err_fd = os.dup(self._err_fd)

            # Create temporary files to redirect the output to. Make sure they
            # aren't opened in binary mode, and specify + for reading and
            # writing.
            self._file_out = tempfile.TemporaryFile(mode='w+')
            self._file_err = tempfile.TemporaryFile(mode='w+')

            # Apply Python-level redirection of sys.stdxxx in addition to the
            # lower-level redirect of sys.__stdxxx___.
            try:
                sys.stdout.flush()
            except AttributeError:  # pragma: no cover
                pass
            finally:
                sys.stdout = self._file_out
            try:
                sys.stderr.flush()
            except AttributeError:  # pragma: no cover
                pass
            finally:
                sys.stderr = self._file_err

            # If possible, overwrite the stdout and stderr file descriptors
            # with the file descriptor for our files. C/C++ code etc. will
            # still look in this location for a descriptor, so that output is
            # redirected.
            if self._out_fd >= 0:
                sys.__stdout__.flush()
                os.dup2(self._file_out.fileno(), self._out_fd)
            if self._err_fd >= 0:
                sys.__stderr__.flush()
                os.dup2(self._file_err.fileno(), self._err_fd)

    def _stop(self):
        """Stops capturing output."""

        # Flush any remaining output
        sys.stdout.flush()
        sys.stderr.flush()

        # Undo redirect
        if not self._fd:
            # Direct back at original
            sys.stdout = self._org_out
            sys.stderr = self._org_err

            # Store text
            self._txt_out = self._tmp_out.getvalue()
            self._txt_err = self._tmp_err.getvalue()

            # Tidy
            self._tmp_out = self._tmp_err = None

        else:
            # Undo redirection
            sys.stdout = self._org_out
            if self._org_out_fd is not None:
                # Copy our backup of the original fd into out_fd
                os.dup2(self._org_out_fd, self._out_fd)
                # And close our backup
                os.close(self._org_out_fd)

            sys.stderr = self._org_err
            if self._org_err_fd is not None:
                os.dup2(self._org_err_fd, self._err_fd)
                os.close(self._org_err_fd)

            # Close temporary files and store capture output
            try:
                self._file_out.seek(0)
                self._txt_out = self._file_out.read()
                self._file_out.close()
            except ValueError:  # pragma: no cover
                # In rare cases, I've seen a ValueError, "underlying buffer has
                # been detached".
                pass
            try:
                self._file_err.seek(0)
                self._txt_err = self._file_err.read()
                self._file_err.close()
            except ValueError:  # pragma: no cover
                pass

            # Tidy
            self._out_fd = self._err_fd = None
            self._org_out_fd = self._org_err_fd = None
            self._file_out = self._file_err = None

        # Unset reference to original streams (for proper decreffing etc.)
        self._stdout = self._stderr = None

    def err(self):
        """
        Returns the text captured from stderr, or an empty string if nothing
        was captured or capturing is still active.
        """
        if self._txt_err is None:
            return ''
        text = self._txt_err

        return text

    def out(self):
        """
        Returns the text captured from stdout, or an empty string if nothing
        was captured or capturing is still active.
        """
        if self._txt_out is None:
            return ''
        text = self._txt_out

        return text

    def text(self):
        """
        Returns the combined text captured from output and error text, if any
        (output first, then error text).
        """
        return self.out() + self.err()


def format_path(path, root='.'):
    """
    Formats a path for use in user messages. If the given path is a
    subdirectory of the current directory this part is chopped off.

    Alternatively, a ``root`` directory may be given explicitly: any
    subdirectory of this path will be formatted relative to ``root``.

    This function differs from os.path.relpath() in the way it handles paths
    *outside* the root: In these cases relpath returns a relative path such as
    '../../' while this function returns an absolute path.
    """
    if path == '':
        path = '.'
    try:
        path = os.path.relpath(path, root)
    except ValueError:  # pragma: no cover
        # This can happen on windows, if `path` is on a different drive than
        # root (so that no relative path from one to the other can be made).
        return path

    if '..' in path:
        path = os.path.abspath(os.path.join(root, path))
    return path


def lvsd(s, t):
    """Returns the Levenshtein distance between two strings ``s`` and ``t``."""
    # This implementation is adapted from wikipedia:
    # en.wikipedia.org/wiki/Levenshtein_distance#Iterative_with_full_matrix
    if s == t:
        return 0
    if len(s) == 0:
        return len(t)
    if len(t) == 0:
        return len(s)

    n = len(t) + 1
    v0 = list(range(n))
    v1 = [None] * n
    for i, si in enumerate(s):
        v1[0] = i + 1
        for j, tj in enumerate(t):
            v1[j + 1] = min(
                v0[j + 1] + 1, v1[j] + 1, v0[j] if si == tj else v0[j] + 1)
        v0, v1 = v1, v0
    return v0[n - 1]


def natural_sort_key(s):
    """
    Function to use as ``key`` in a sort, to get natural sorting of strings
    (e.g. "2" before "10").

    Usage examples::

        names.sort(key=myokit.tools.natural_sort_key)

        variables.sort(key=lambda v: myokit.tools.natural_sort_key(v.qname()))

    """
    # Code adapted from: http://stackoverflow.com/questions/4836710/
    return [
        int(text) if text.isdigit() else text.lower()
        for text in _natural_sort_regex.split(s)]


def rmtree(path, silent=False):
    """
    Version of ``shutil.rmtree`` that handles Windows "access denied" errors
    (when the user is lacking write permissions, but is allowed to set them).

    If ``silent=True`` any other exceptions will be caught and ignored.
    """
    # From https://stackoverflow.com/questions/2656322
    def onerror(function, path, excinfo):   # pragma: no cover
        if not os.access(path, os.W_OK):
            # Give user write permissions (remove read-only flag)
            os.chmod(path, stat.S_IWUSR)
            function(path)
        else:
            raise

    if silent:
        try:
            shutil.rmtree(path, ignore_errors=False, onerror=onerror)
        except Exception:
            pass
    else:
        shutil.rmtree(path, ignore_errors=False, onerror=onerror)




#
# Exports to plain python
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os

import myokit.formats


class PythonExporter(myokit.formats.TemplatedRunnableExporter):
    """
    This :class:`Exporter <myokit.formats.Exporter>` generates a python-only
    model implementation and a simple (but slow) integration routine.

    Both the model definition and the pacing protocol are fully exported.

    Provides the following external variables:

    ``time``
        The current simulation time
    ``pace``
        The current value of the pacing system, implemented using the given
        protocol.

    A graph of the results is generated using matplotlib.
    """

    def _dir(self, root):
        return os.path.join(root, 'python', 'template')

    def _dict(self):
        return {'sim.py': 'sim.py'}

    def _vars(self, model, protocol):
        return {'model': model, 'protocol': protocol}



#
# Widget to display graphs of variable's equations.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit
import myokit.gui

from myokit.gui import QtGui, QtWidgets

import numpy as np

# Matplotlib (must be imported _after_ gui has had chance to set backend)
import matplotlib
import matplotlib.figure
from myokit.gui import matplotlib_backend as backend
from mpl_toolkits.mplot3d import axes3d     # noqa


# Constants
# Classes & methods
class VarGrapher(QtWidgets.QDialog):
    """
    This widget is used to display graphs of model variables.

    Arguments:

    ``parent``
        The parent window calling this one.
    ``title``
        A title for this window
    ``var``
        A :class:`myokit.Variable`.
    ``func``
        The python function that calculates ``var``
    ``args``
        The arguments to func, i.e. the LhsExpressions that ``var`` depends on.

    *Extends:* ``QtWidgets.QDialog``
    """
    def __init__(self, parent, title, var, func, args):
        super().__init__(parent)
        self.setFixedSize(700, 600)
        self.setWindowTitle(title)

        # Figure panel
        self._figure = matplotlib.figure.Figure()
        self._canvas = backend.FigureCanvasQTAgg(self._figure)
        self._toolbar = backend.NavigationToolbar2QT(self._canvas, self)

        # Variable panel
        self._variable_widget = QtWidgets.QWidget()

        # Button panel
        self._button_widget = QtWidgets.QWidget()

        # Central widget
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self._canvas)
        layout.addWidget(self._toolbar)
        layout.addWidget(self._variable_widget)
        layout.addWidget(self._button_widget)
        self.setLayout(layout)

        # Get function handle, information object
        self._func = func
        self._args = args
        self._var = var

        # Variable ranges
        grid = QtWidgets.QGridLayout()
        self._bounds = {}
        for k, lhs in enumerate(self._args):
            var = lhs.var()
            # Guess appropriate bounds
            if var.label() == 'membrane_potential' or \
                    var.name().lower() in ['v', 'voltage', 'potential']:
                if var.unit() == myokit.units.volt:
                    lohi = (-0.1, 0.1)
                else:
                    lohi = (-100.0, 100.0)
            else:
                v = lhs.eval()
                if v >= 0 and v <= 1:
                    lohi = (0.0, 1.0)
                elif v < 0:
                    lohi = (-50, 50)
                else:
                    lohi = (0, 100)
            # Row and column of first widget in grid
            row = k // 2
            col = (k % 2) * 3
            # Add label
            label = QtWidgets.QLabel(var.qname())
            grid.addWidget(label, row, col)
            # Add lower and upper bound or single value
            if k < 2:
                # Lower
                editlo = QtWidgets.QLineEdit()
                editlo.setValidator(QtGui.QDoubleValidator())
                editlo.setText(str(lohi[0]))
                grid.addWidget(editlo, row, col + 1)
                # Upper
                edithi = QtWidgets.QLineEdit()
                edithi.setValidator(QtGui.QDoubleValidator())
                edithi.setText(str(lohi[1]))
                grid.addWidget(edithi, row, col + 2)
                self._bounds[lhs] = (editlo, edithi)
            else:
                # Single, fixed value
                v = 0.5 * (lohi[0] + lohi[1])
                edit = QtWidgets.QLineEdit(str(v))
                edit.setReadOnly(True)
                grid.addWidget(edit, row, col + 1)
                self._bounds[lhs] = (edit, edit)
        self._variable_widget.setLayout(grid)

        # Buttons
        layout = QtWidgets.QHBoxLayout()

        # Graph button
        button = QtWidgets.QPushButton('Refresh')
        button.clicked.connect(self.action_draw)
        layout.addWidget(button)

        # Close button
        button = QtWidgets.QPushButton('Close')
        button.clicked.connect(self.close)
        layout.addWidget(button)
        self._button_widget.setLayout(layout)

        # Draw!
        self.action_draw()

    def action_draw(self):
        """
        (re-)draws the graph
        """
        n = len(self._args)
        if n == 0:
            raise Exception('Cannot graph constant')
        elif n == 1:
            steps = 10000
        else:
            steps = 100

        # Get boundaries
        vrs = []
        for arg in self._args:
            lohi = self._bounds[arg]
            lo = float(lohi[0].text())
            hi = float(lohi[1].text())
            if hi < lo:
                lo, hi = hi, lo
            if lo == hi:
                var = lo
            else:
                var = np.linspace(lo, hi, steps)
            vrs.append(var)

        def label(lhs):
            label = str(lhs)
            unit = lhs.var().unit()
            return label if unit is None else label + ' ' + str(unit)

        # Redraw figure
        try:
            self._figure.clear()
        except Exception:
            pass

        if n == 1:
            ax = self._figure.add_subplot(111)
            x = vrs[0]
            with np.errstate(all='ignore'):
                y = self._func(x)
            ax.plot(x, y)
            ax.grid(True)
            ax.set_xlabel(label(self._args[0]))
            ax.set_ylabel(label(self._var.lhs()))
        else:
            ax = self._figure.add_subplot(111, projection='3d')
            x = vrs[0]
            y = vrs[1]
            x, y = np.meshgrid(x, y)
            vrs[0] = x
            vrs[1] = y
            with np.errstate(all='ignore'):
                z = self._func(*vrs)
            ax.plot_surface(x, y, z)
            ax.grid(True)
            ax.set_xlabel(label(self._args[0]))
            ax.set_ylabel(label(self._args[1]))
            ax.set_zlabel(label(self._var.lhs()))

        # Refresh
        self._canvas.draw()
        self.update()
